#!/usr/bin/env amspython
# coding: utf-8

# ## Initial imports

from scm.plams import *
from ase.visualize.plot import plot_atoms
from ase.build import fcc111, bulk
import matplotlib.pyplot as plt


# ## Helper functions


def printsummary(mol, details=None):
    if details:
        density = details["density"]
    else:
        density = mol.get_density() * 1e-3
    s = f"{len(mol)} atoms, density = {density:.3f} g/cm^3"
    if mol.lattice:
        s += f", box = {mol.lattice[0][0]:.3f}, {mol.lattice[1][1]:.3f}, {mol.lattice[2][2]:.3f}"
    s += f", formula = {mol.get_formula()}"
    if details:
        s += f'\n#added molecules per species: {details["n_molecules"]}, mole fractions: {details["mole_fractions"]}'
    print(s)


def show(mol, figsize=None, **kwargs):
    """Show a molecule in a Jupyter notebook"""
    plt.figure(figsize=figsize or (2, 2))
    plt.axis("off")
    plot_atoms(toASE(mol), **kwargs)


# ## Liquid water (fluid with 1 component)
# First, create the gasphase molecule:

water = from_smiles("O")
show(water)


print("pure liquid from approximate number of atoms and exact density (in g/cm^3), cubic box with auto-determined size")
out = packmol(water, n_atoms=194, density=1.0)
printsummary(out)
out.write("water-1.xyz")
show(out)


print("pure liquid from approximate density (in g/cm^3) and an orthorhombic box")
out = packmol(water, density=1.0, box_bounds=[0.0, 0.0, 0.0, 8.0, 12.0, 14.0])
printsummary(out)
out.write("water-2.xyz")
show(out)


print("pure liquid with explicit number of molecules and exact density")
out = packmol(water, n_molecules=64, density=1.0)
printsummary(out)
out.write("water-3.xyz")
show(out)


print("pure liquid with explicit number of molecules and box")
out = packmol(water, n_molecules=64, box_bounds=[0.0, 0.0, 0.0, 12.0, 13.0, 14.0])
printsummary(out)
out.write("water-4.xyz")
show(out)


print("water-5.xyz: pure liquid in non-orthorhombic box (requires AMS2022 or later)")
# first place the molecules in a cuboid surrounding the desired lattice
# then gradually change into the desired lattice using refine_lattice()
# note that the molecules may become distorted by this procedure
lattice = [[10.0, 2.0, -1.0], [-5.0, 8.0, 0.0], [0.0, -2.0, 11.0]]
temp_out = packmol(
    water,
    n_molecules=32,
    box_bounds=[
        0,
        0,
        0,
        max(lattice[i][0] for i in range(3)) - min(lattice[i][0] for i in range(3)),
        max(lattice[i][1] for i in range(3)) - min(lattice[i][1] for i in range(3)),
        max(lattice[i][2] for i in range(3)) - min(lattice[i][2] for i in range(3)),
    ],
)
out = refine_lattice(temp_out, lattice=lattice)
if out is not None:
    out.write("water-5.xyz")
    print(
        "Top: system in surrounding orthorhombic box before calling refine_lattice(). Bottom: System in non-orthorhombic box after calling refine_lattice()"
    )
    show(temp_out)
    show(out)


# ## Water-acetonitrile mixture (fluid with 2 or more components)
# Let's also create a single acetonitrile molecule:

acetonitrile = from_smiles("CC#N")
show(acetonitrile)


# Set the desired mole fractions and density. Here, the density is calculated as the weighted average of water (1.0 g/cm^3) and acetonitrile (0.76 g/cm^3) densities, but you could use any other density.

# MIXTURES
x_water = 0.666  # mole fraction
x_acetonitrile = 1 - x_water  # mole fraction
density = (x_water * 1.0 + x_acetonitrile * 0.76) / (
    x_water + x_acetonitrile
)  # weighted average of pure component densities

print("MIXTURES")
print(f"x_water = {x_water:.3f}")
print(f"x_acetonitrile = {x_acetonitrile:.3f}")
print(f"target density = {density:.3f} g/cm^3")


# By setting ``return_details=True``, you can get information about the mole fractions of the returned system. They may not exactly match the mole fractions you put in.

print(
    "2-1 water-acetonitrile from approximate number of atoms and exact density (in g/cm^3), cubic box with auto-determined size"
)
out, details = packmol(
    molecules=[water, acetonitrile],
    mole_fractions=[x_water, x_acetonitrile],
    n_atoms=200,
    density=density,
    return_details=True,
)
printsummary(out, details)
out.write("water-acetonitrile-1.xyz")
show(out)


# The ``details`` is a dictionary as follows:

for k, v in details.items():
    print(f"{k}: {v}")


print("2-1 water-acetonitrile from approximate density (in g/cm^3) and box bounds")
out, details = packmol(
    molecules=[water, acetonitrile],
    mole_fractions=[x_water, x_acetonitrile],
    box_bounds=[0, 0, 0, 13.2, 13.2, 13.2],
    density=density,
    return_details=True,
)
printsummary(out, details)
out.write("water-acetonitrile-2.xyz")
show(out)


print("2-1 water-acetonitrile from explicit number of molecules and density, cubic box with auto-determined size")
out, details = packmol(molecules=[water, acetonitrile], n_molecules=[32, 16], density=density, return_details=True)
printsummary(out, details)
out.write("water-acetonitrile-3.xyz")
show(out)


print("2-1 water-acetonitrile from explicit number of molecules and box")
out = packmol(molecules=[water, acetonitrile], n_molecules=[32, 16], box_bounds=[0, 0, 0, 13.2, 13.2, 13.2])
printsummary(out)
out.write("water-acetonitrile-4.xyz")
show(out)


# ## Pack inside sphere
#
# Set ``sphere=True`` to pack in a sphere (non-periodic) instead of in a periodic box. The sphere will be centered near the origin.

print("water in a sphere from exact density and number of molecules")
out, details = packmol(molecules=[water], n_molecules=[100], density=1.0, return_details=True, sphere=True)
printsummary(out, details)
print(f"Radius  of sphere: {details['radius']:.3f} ang.")
print(f"Center of mass xyz (ang): {out.get_center_of_mass()}")
out.write("water-sphere.xyz")
show(out)


print(
    "2-1 water-acetonitrile in a sphere from exact density (in g/cm^3) and approximate number of atoms and mole fractions"
)
out, details = packmol(
    molecules=[water, acetonitrile],
    mole_fractions=[x_water, x_acetonitrile],
    n_atoms=500,
    density=density,
    return_details=True,
    sphere=True,
)
printsummary(out, details)
out.write("water-acetonitrile-sphere.xyz")
show(out)


# ## Packing ions, total system charge
#
# The total system charge will be sum of the charges of the constituent molecules.
#
# In PLAMS, ``molecule.properties.charge`` specifies the charge:

ammonium = from_smiles("[NH4+]")  # ammonia.properties.charge == +1
chloride = from_smiles("[Cl-]")  # chloride.properties.charge == -1
print("3 water molecules, 3 ammonium, 1 chloride (non-periodic)")
print("Initial charges:")
print(f"Water: {water.properties.get('charge', 0)}")
print(f"Ammonia: {ammonium.properties.get('charge', 0)}")
print(f"Chloride: {chloride.properties.get('charge', 0)}")
out = packmol(molecules=[water, ammonium, chloride], n_molecules=[3, 3, 1], density=0.4, sphere=True)
tot_charge = out.properties.get("charge", 0)
print(f"Total charge of packmol-generated system: {tot_charge}")
out.write("water-ammonium-chloride.xyz")
show(out)


# ## Microsolvation
# ``packmol_microsolvation`` can create a microsolvation sphere around a solute.

out = packmol_microsolvation(solute=acetonitrile, solvent=water, density=1.5, threshold=4.0)
# for microsolvation it's a good idea to have a higher density than normal to get enough solvent molecules
print(f"Microsolvated structure: {len(out)} atoms.")
out.write("acetonitrile-microsolvated.xyz")

figsize = (3, 3)
show(out, figsize=figsize)


# ## Solid-liquid or solid-gas interfaces
# First, create a slab using the ASE ``fcc111`` function

rotation = "90x,0y,0z"  # sideview of slab
slab = fromASE(fcc111("Al", size=(4, 6, 3), vacuum=15.0, orthogonal=True, periodic=True))
show(slab, figsize=figsize, rotation=rotation)


print("water surrounding an Al slab, from an approximate density")
out = packmol_on_slab(slab, water, density=1.0)
printsummary(out)
out.write("al-water-pure.xyz")
show(out, figsize=figsize, rotation=rotation)


print("2-1 water-acetonitrile mixture surrounding an Al slab, from mole fractions and an approximate density")
out = packmol_on_slab(slab, [water, acetonitrile], mole_fractions=[x_water, x_acetonitrile], density=density)
printsummary(out)
out.write("al-water-acetonitrile.xyz")
show(out, figsize=figsize, rotation=rotation)


# ## Pack inside voids in crystals
#
# Use the ``packmol_in_void`` function. You can decrease ``tolerance`` if you need to pack very tightly. The default value for ``tolerance`` is 2.0.

bulk_Al = fromASE(bulk("Al", cubic=True).repeat((3, 3, 3)))
rotation = "90x,5y,5z"
show(bulk_Al, rotation=rotation, radii=0.4)


out = packmol_in_void(
    host=bulk_Al, molecules=[from_smiles("[H]"), from_smiles("[He]")], n_molecules=[50, 20], tolerance=1.5
)
show(out, rotation=rotation, radii=0.4)
printsummary(out)
out.write("al-bulk-with-h-he.xyz")


# ## Bonds, atom properties (force field types, regions, ...)
#
# The ``packmol()`` function accepts the arguments ``keep_bonds`` and ``keep_atom_properties``. These options will keep the bonds defined for the constitutent molecules, as well as any atomic properties.
#
# The bonds and atom properties are easiest to see by printing the System block for an AMS job:

water = from_smiles("O")
n2 = from_smiles("N#N")

# delete properties coming from from_smiles
for at in water:
    at.properties = Settings()
for at in n2:
    at.properties = Settings()

water[1].properties.region = "oxygen_atom"
water[2].properties.mass = 2.014  # deuterium
water.delete_bond(water[1, 2])  # delete bond between atoms 1 and 2 (O and H)


out = packmol([water, n2], n_molecules=[2, 1], density=0.5)
print(AMSJob(molecule=out).get_input())


# By default, the ``packmol()`` function assigns regions called ``mol0``, ``mol1``, etc. to the different added molecules. The ``region_names`` option lets you set custom names.

out = packmol([water, n2], n_molecules=[2, 1], density=0.5, region_names=["water", "nitrogen_molecule"])
print(AMSJob(molecule=out).get_input())


# Below, we also set ``keep_atom_properties=False``, this will remove the previous regions (in this example "oxygen_atom") and mass.

out = packmol([water, n2], n_molecules=[2, 1], density=0.5, keep_atom_properties=False)
print(AMSJob(molecule=out).get_input())


# ``keep_bonds=False`` will additionally ignore any defined bonds:

out = packmol(
    [water, n2],
    n_molecules=[2, 1],
    density=0.5,
    region_names=["water", "nitrogen_molecule"],
    keep_bonds=False,
    keep_atom_properties=False,
)
print(AMSJob(molecule=out).get_input())



from scm.plams.version import __version__
from scm.plams.core.basejob import MultiJob, SingleJob
from scm.plams.core.errors import (
    FileError,
    JobError,
    MoleculeError,
    PlamsError,
    PTError,
    ResultsError,
    TrajectoryError,
    UnitsError,
)
from scm.plams.core.functions import (
    add_to_class,
    add_to_instance,
    config,
    delete_job,
    finish,
    init,
    load,
    load_all,
    log,
    read_all_molecules_in_xyz_file,
    read_molecules,
)
from scm.plams.core.jobmanager import JobManager
from scm.plams.core.jobrunner import GridRunner, JobRunner
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.adf import ADFJob, ADFResults
from scm.plams.interfaces.adfsuite.ams import AMSJob, AMSResults
from scm.plams.interfaces.adfsuite.amsanalysis import (
    AMSAnalysisJob,
    AMSAnalysisResults,
    convert_to_unicode,
)
from scm.plams.interfaces.adfsuite.amspipeerror import (
    AMSPipeDecodeError,
    AMSPipeError,
    AMSPipeInvalidArgumentError,
    AMSPipeLogicError,
    AMSPipeRuntimeError,
    AMSPipeUnknownArgumentError,
    AMSPipeUnknownMethodError,
    AMSPipeUnknownVersionError,
)
from scm.plams.interfaces.adfsuite.amsworker import (
    AMSWorker,
    AMSWorkerError,
    AMSWorkerPool,
    AMSWorkerResults,
)
from scm.plams.interfaces.adfsuite.band import BANDJob, BANDResults
from scm.plams.interfaces.adfsuite.crs import CRSJob, CRSResults
from scm.plams.interfaces.adfsuite.densf import DensfJob, DensfResults
from scm.plams.interfaces.adfsuite.dftb import DFTBJob, DFTBResults
from scm.plams.interfaces.adfsuite.fcf import FCFJob, FCFResults
from scm.plams.interfaces.adfsuite.forcefieldparams import (
    ForceFieldPatch,
    forcefield_params_from_kf,
)
from scm.plams.interfaces.adfsuite.mopac import MOPACJob, MOPACResults
from scm.plams.interfaces.adfsuite.quickjobs import (
    preoptimize,
    refine_density,
    refine_lattice,
)
from scm.plams.interfaces.adfsuite.reaxff import (
    ReaxFFJob,
    ReaxFFResults,
    load_reaxff_control,
    reaxff_control_to_settings,
)
from scm.plams.interfaces.adfsuite.uff import UFFJob, UFFResults
from scm.plams.interfaces.adfsuite.unifac import UnifacJob, UnifacResults
from scm.plams.interfaces.molecule.ase import fromASE, toASE
from scm.plams.interfaces.molecule.packmol import (
    PackMolError,
    packmol,
    packmol_microsolvation,
    packmol_on_slab,
    packmol_in_void,
)
from scm.plams.interfaces.molecule.rdkit import (
    add_Hs,
    apply_reaction_smarts,
    apply_template,
    canonicalize_mol,
    from_rdmol,
    from_sequence,
    from_smarts,
    from_smiles,
    gen_coords_rdmol,
    get_backbone_atoms,
    get_conformations,
    get_substructure,
    modify_atom,
    partition_protein,
    readpdb,
    to_rdmol,
    to_smiles,
    writepdb,
    yield_coords,
)
from scm.plams.interfaces.thirdparty.cp2k import Cp2kJob, Cp2kResults, Cp2kSettings2Mol
from scm.plams.interfaces.thirdparty.crystal import CrystalJob, mol2CrystalConf
from scm.plams.interfaces.thirdparty.dftbplus import DFTBPlusJob, DFTBPlusResults
from scm.plams.interfaces.thirdparty.dirac import DiracJob, DiracResults
from scm.plams.interfaces.thirdparty.gamess import GamessJob
from scm.plams.interfaces.thirdparty.orca import ORCAJob, ORCAResults
from scm.plams.interfaces.thirdparty.raspa import RaspaJob, RaspaResults
from scm.plams.interfaces.thirdparty.vasp import VASPJob, VASPResults
from scm.plams.mol.atom import Atom
from scm.plams.mol.bond import Bond
from scm.plams.mol.identify import label_atoms
from scm.plams.mol.molecule import Molecule
from scm.plams.mol.pdbtools import PDBHandler, PDBRecord
from scm.plams.recipes.adffragment import ADFFragmentJob, ADFFragmentResults
from scm.plams.recipes.adfnbo import ADFNBOJob
from scm.plams.recipes.md.amsmdjob import AMSMDJob, AMSNVEJob, AMSNVTJob, AMSNPTJob
from scm.plams.recipes.md.nvespawner import AMSNVESpawnerJob
from scm.plams.recipes.md.trajectoryanalysis import AMSRDFJob, AMSVACFJob, AMSMSDJob
from scm.plams.recipes.md.scandensity import AMSMDScanDensityJob
from scm.plams.recipes.numgrad import NumGradJob
from scm.plams.recipes.numhess import NumHessJob
from scm.plams.recipes.pestools.optimizer import Optimizer
from scm.plams.recipes.reorganization_energy import ReorganizationEnergyJob
from scm.plams.recipes.redox import AMSRedoxDirectJob, AMSRedoxScreeningJob, AMSRedoxThermodynamicCycleJob

from scm.plams.tools.converters import (
    file_to_traj,
    gaussian_output_to_ams,
    qe_output_to_ams,
    rkf_to_ase_atoms,
    rkf_to_ase_traj,
    traj_to_rkf,
    vasp_output_to_ams,
)
from scm.plams.tools.geometry import (
    angle,
    axis_rotation_matrix,
    cell_shape,
    cellvectors_from_shape,
    dihedral,
    distance_array,
    rotation_matrix,
)
from scm.plams.tools.kftools import KFFile, KFHistory, KFReader
from scm.plams.tools.periodic_table import PT, PeriodicTable
from scm.plams.tools.plot import plot_band_structure, plot_molecule
from scm.plams.tools.reaction_energies import (
    balance_equation,
    get_stoichiometry,
    reaction_energy,
)
from scm.plams.tools.units import Units
from scm.plams.trajectories.dcdfile import DCDTrajectoryFile
from scm.plams.trajectories.rkffile import (
    RKFTrajectoryFile,
    write_general_section,
    write_molecule_section,
)
from scm.plams.trajectories.rkfhistoryfile import (
    RKFHistoryFile,
    molecules_to_rkf,
    rkf_filter_regions,
)
from scm.plams.trajectories.sdffile import SDFTrajectoryFile, create_sdf_string
from scm.plams.trajectories.sdfhistoryfile import SDFHistoryFile
from scm.plams.trajectories.trajectory import Trajectory
from scm.plams.trajectories.trajectoryfile import TrajectoryFile
from scm.plams.trajectories.xyzfile import XYZTrajectoryFile, create_xyz_string
from scm.plams.trajectories.xyzhistoryfile import XYZHistoryFile

__all__ = [
    "Results",
    "JobRunner",
    "GridRunner",
    "init",
    "finish",
    "log",
    "load",
    "load_all",
    "delete_job",
    "add_to_class",
    "add_to_instance",
    "config",
    "read_molecules",
    "read_all_molecules_in_xyz_file",
    "JobManager",
    "Settings",
    "SingleJob",
    "MultiJob",
    "PlamsError",
    "FileError",
    "ResultsError",
    "JobError",
    "PTError",
    "UnitsError",
    "MoleculeError",
    "TrajectoryError",
    "Bond",
    "label_atoms",
    "PDBRecord",
    "PDBHandler",
    "Molecule",
    "Atom",
    "GamessJob",
    "DiracJob",
    "DiracResults",
    "RaspaJob",
    "RaspaResults",
    "CrystalJob",
    "mol2CrystalConf",
    "DFTBPlusJob",
    "DFTBPlusResults",
    "Cp2kJob",
    "Cp2kResults",
    "Cp2kSettings2Mol",
    "VASPJob",
    "VASPResults",
    "ORCAJob",
    "ORCAResults",
    "DFTBJob",
    "DFTBResults",
    "BANDJob",
    "BANDResults",
    "UFFJob",
    "UFFResults",
    "CRSResults",
    "CRSJob",
    "AMSPipeError",
    "AMSPipeDecodeError",
    "AMSPipeLogicError",
    "AMSPipeRuntimeError",
    "AMSPipeUnknownVersionError",
    "AMSPipeUnknownMethodError",
    "AMSPipeUnknownArgumentError",
    "AMSPipeInvalidArgumentError",
    "ForceFieldPatch",
    "forcefield_params_from_kf",
    "AMSWorker",
    "AMSWorkerResults",
    "AMSWorkerError",
    "AMSWorkerPool",
    "DensfJob",
    "DensfResults",
    "ADFJob",
    "ADFResults",
    "FCFJob",
    "FCFResults",
    "MOPACJob",
    "MOPACResults",
    "AMSAnalysisJob",
    "AMSAnalysisResults",
    "convert_to_unicode",
    "preoptimize",
    "refine_density",
    "refine_lattice",
    "ReaxFFJob",
    "ReaxFFResults",
    "load_reaxff_control",
    "reaxff_control_to_settings",
    "UnifacJob",
    "UnifacResults",
    "AMSJob",
    "AMSResults",
    "toASE",
    "fromASE",
    "packmol",
    "packmol_on_slab",
    "packmol_microsolvation",
    "packmol_in_void",
    "PackMolError",
    "add_Hs",
    "apply_reaction_smarts",
    "apply_template",
    "gen_coords_rdmol",
    "get_backbone_atoms",
    "modify_atom",
    "to_rdmol",
    "from_rdmol",
    "from_sequence",
    "from_smiles",
    "from_smarts",
    "to_smiles",
    "partition_protein",
    "readpdb",
    "writepdb",
    "get_substructure",
    "get_conformations",
    "yield_coords",
    "canonicalize_mol",
    "KFFile",
    "KFReader",
    "KFHistory",
    "traj_to_rkf",
    "vasp_output_to_ams",
    "qe_output_to_ams",
    "gaussian_output_to_ams",
    "rkf_to_ase_traj",
    "rkf_to_ase_atoms",
    "file_to_traj",
    "Units",
    "rotation_matrix",
    "axis_rotation_matrix",
    "distance_array",
    "angle",
    "dihedral",
    "cell_shape",
    "cellvectors_from_shape",
    "get_stoichiometry",
    "balance_equation",
    "reaction_energy",
    "PeriodicTable",
    "PT",
    "plot_band_structure",
    "plot_molecule",
    "SDFTrajectoryFile",
    "create_sdf_string",
    "XYZHistoryFile",
    "Trajectory",
    "TrajectoryFile",
    "RKFTrajectoryFile",
    "write_general_section",
    "write_molecule_section",
    "DCDTrajectoryFile",
    "RKFHistoryFile",
    "molecules_to_rkf",
    "rkf_filter_regions",
    "XYZTrajectoryFile",
    "create_xyz_string",
    "SDFHistoryFile",
    "ReorganizationEnergyJob",
    "ADFFragmentJob",
    "ADFFragmentResults",
    "ADFNBOJob",
    "AMSMDJob",
    "AMSNVEJob",
    "AMSNVTJob",
    "AMSNPTJob",
    "AMSNVESpawnerJob",
    "AMSMDScanDensityJob",
    "AMSRDFJob",
    "AMSMSDJob",
    "AMSVACFJob",
    "NumGradJob",
    "NumHessJob",
    "Optimizer",
    "AMSRedoxDirectJob",
    "AMSRedoxScreeningJob",
    "AMSRedoxThermodynamicCycleJob",
]



import numpy as np

__all__ = ["autocorrelation", "power_spectrum"]


def autocorrelation(np_data, max_dt=None, normalize=False):
    """
    Calculate an autocorrelation function

    np_data : numpy array of shape (n_time_steps, n_data_per_time_step)
        Data for which to calculate the autocorrelation function
    max_dt : int
        Smallest integer delta_time for which the correlation function is not calculated. If max_dt is greater than n_time_steps, n_time_steps is used instead. If None, the largest possible value is used.
    normalize : bool
        Whether to return a normalized autocorrelation function (where the first value is 1)

    Returns : A numpy array of shape (max_dt,) or (max_dt, n_data_per_time_step) (depending on ``average``).
    """
    num_timesteps = np_data.shape[0]
    if max_dt is None or max_dt > num_timesteps:
        max_dt = num_timesteps
    C = np.zeros(max_dt)
    C[0] = 1
    for dt in range(max_dt):
        try:
            Va = np_data[: num_timesteps - dt, :]  # the first num_timesteps-dt rows
            Vb = np_data[dt:, :]  # ignore the first dt rows
            C[dt] = np.mean(Va * Vb)
        except KeyboardInterrupt:
            print("Stopping at dt = {} of max_dt = {}".format(dt, max_dt))
            break
    if normalize:
        C = C / C[0]
    return C


def power_spectrum(times, C, max_freq=None, number_of_points=None):
    """
    Calculate power spectrum from an autocorrelation function
    times : 1D numpy array
        times (in femtoseconds), same length as C. Only the delta between the first two points is needed.
    C : numpy array
        Contains the autocorrelation function
    max_freq: float
        the maximum frequency (in cm-1) to be output
    number_of_points: int
        number of points to include in the FFT.

    Returns: a 2-tuple
        an (x, y) tuple where x = frequency in cm-1, and y = power spectrum as 1D numpy arrays only the real part of the power spectrum is returned
    """
    speedoflight = 29979245800
    # in cm/s
    dt = (times[1] - times[0]) * 1e-15  # in seconds
    max_freq = max_freq or 5000
    if number_of_points is None:
        number_of_points = int(
            1 / (dt * speedoflight)
        )  # the resulting spectrum_x will have points spaced by about 1 cm-1
        if number_of_points < C.size:
            number_of_points = C.size
    spectrum_y = np.fft.fft(C, number_of_points)
    spectrum_x = np.arange(number_of_points) / (dt * speedoflight * number_of_points)
    spectrum_y = np.real(spectrum_y[spectrum_x < max_freq])
    spectrum_x = spectrum_x[spectrum_x < max_freq]
    return spectrum_x.ravel(), spectrum_y.ravel()



#!/usr/bin/env python

import datetime
import os
import threading
import time

from scm.plams.core.errors import FileError, PlamsError
from scm.plams.core.functions import log
from scm.plams.core.jobrunner import JobRunner
from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.interfaces.adfsuite.amsworker import AMSWorkerPool

__all__ = ["Optimizer"]


class Optimizer:
    """
    Machine that can optimize the geometry of a set of ACE molecules
    """

    def __init__(self, nproc=1, jobrunner=None):
        """
        Initiates an instance of the Optimizer class

        * ``nproc``           -- Number of processors to use in total for the optimizations
        """
        # The general job settings
        self.settings = Settings()
        self.settings.runscript.nproc = nproc

        # The settings for geometry optimizations
        self.go_settings = Settings()
        self.go_settings.input.ams.GeometryOptimization.Method = "Quasi-Newton"
        self.go_settings.input.ams.GeometryOptimization.CoordinateType = "Auto"  # Cartesian?
        # self.method = 'Quasi-Newton'
        # self.coordinate_type = 'Auto' # Cartesian?

        # Set the jobrunner
        if jobrunner is None:
            parallel = False
            if nproc > 1:
                parallel = True
            jobrunner = JobRunner(parallel=parallel, maxjobs=nproc)
        self.jobrunner = jobrunner

        # Wether or not to keep the files after running
        self.keep = None  # 'all' is the PLAMS default
        self.use_pipe = True
        self.last_timings = None
        self.watch = False

        # Private variables
        self.path = None
        self.nsteps = None

    def set_cleaning_preference(self, keep=None):
        """
        Sets the preferences for the files saved to disc

        * ``keep`` -- None or 'all'.
         The former will clean PLAMS folders of all files after runtime,
         the latter will leave all files as they are.
         See PLAMS for more details.
        """
        if keep is not None:
            self.use_pipe = False
        self.keep = keep

    def keep_ams_running(self, keep=True):
        """
        Technical setting for a set of (serial) geometry optimizations

        * ``keep`` -- Each of the parallel processes will most likely be passed multiple geometries.
         If `keep=True`, AMS will be kept running in the background, and geometries will be passed via a pipe.
         If `keep=False`, each geometry optimization will start a new AMS process.
        """
        self.use_pipe = keep

    def optimize_geometries(self, molecules, engine_settings=None, name="go"):
        """
        Optimize a set of molecules

        * ``engine_settings`` -- PLAMS Settings object:
                    engine_settings = Settings()
                    engine_settings.DFTB.Model = 'GFN1-xTB'
        """
        if not self.use_pipe or len(molecules) == 1:
            molecules, energies, resultlist = self._perform_tasks_separately(molecules, engine_settings, name)
        else:
            molecules, energies, resultlist = self._perform_tasks_using_pipe(molecules, engine_settings, name)
        remaining_energies = [en for en in energies if not en is None]
        log("%s out of %s optimizations succeeded" % (len(remaining_energies), len(energies)))
        return molecules, energies

    def compute_energies(self, molecules, engine_settings=None, name="sp"):
        """
        Compute the energies for a set of molecules

        * ``engine_settings`` -- PLAMS Settings object:
                    engine_settings = Settings()
                    engine_settings.DFTB.Model = 'GFN1-xTB'
        """
        if not self.use_pipe or len(molecules) == 1:
            molecules, energies, resultlist = self._perform_tasks_separately(
                molecules, engine_settings, name, "SinglePoint"
            )
        else:
            molecules, energies, resultlist = self._perform_tasks_using_pipe(
                molecules, engine_settings, name, "SinglePoint"
            )
        remaining_energies = [en for en in energies if not en is None]
        log("%s out of %s single point calculations succeeded" % (len(remaining_energies), len(energies)))
        return molecules, energies

    def _perform_tasks_separately(self, molecules, engine_settings=None, name="go", task="GeometryOptimization"):
        """
        Optimize a set of molecules
        """
        # Set the settings object
        if task == "GeometryOptimization":
            settings = self.get_go_settings()
        elif task == "SinglePoint":
            settings = self.get_sp_settings()
        else:
            raise PlamsError("Task not implemented: %s" % (task))

        # Add the engine settings
        if engine_settings is None:
            engine_settings = Settings()
            engine_settings.ForceField.Type = "UFF"
        engine_name = list(engine_settings.keys())[0]
        settings.input[engine_name] = engine_settings[engine_name]

        # Adjust the number of processes if thre is only one job!
        if len(molecules) > 0:
            settings.runscript.nproc = int(self.jobrunner.maxjobs / min(self.jobrunner.maxjobs, len(molecules)))

        # Run jobs
        taskmap = {"GeometryOptimization": "geometry optimization", "SinglePoint": "single point"}
        starttime = time.time()
        resultlist = []
        if self.watch:
            # Set up and start the progress monitor.
            done_event = threading.Event()
            args = (resultlist, len(molecules), done_event, 60)
            pmt = threading.Thread(target=progress_monitor, args=args)
            log(f"Running {len(molecules)} {taskmap[task]}s")
            pmt.start()

        # Run jobs
        for i, mol in enumerate(molecules):
            job = AMSJob(name="%s%i" % (name, i), molecule=mol, settings=settings)
            resultlist.append(job.run(jobrunner=self.jobrunner))

        # Read results
        if task == "GeometryOptimization":
            optimized_geometries, energies = self._read_results_go(resultlist)
        elif task == "SinglePoint":
            optimized_geometries, energies = self._read_results_sp(resultlist)
        endtime = time.time()
        self.last_timings = endtime - starttime

        # Stop the progreess monitor (all the jobs have finished)
        if self.watch:
            done_event.set()
            pmt.join()
            time_taken = datetime.timedelta(seconds=round(time.time() - starttime))
            txt = f"All {len(molecules)} {taskmap[task]}s done! Time taken: {time_taken}s"
            log(txt)

        # Place the new coordinates into the molecules
        for i, (crd, en) in enumerate(zip(optimized_geometries, energies)):
            if en is not None:
                if crd is not None:
                    molecules[i].from_array(crd)
            molecules[i].properties.energy = en

        # Remove files
        keep = self.keep
        # if len(molecules) == 1 or task == 'SinglePoint' : keep = 'all'
        if len(molecules) == 1:
            keep = "all"
        for i, r in enumerate(resultlist):
            r._clean(keep)
            if keep is None:
                r.job.jobmanager.remove_job(r.job)

        return molecules, energies, resultlist

    def _perform_tasks_using_pipe(self, molecules, engine_settings=None, name="go", task="GeometryOptimization"):
        """
        Optimize a set of ACE molecules
        """
        if len(molecules) == 0:
            return [], [], []

        # Add the engine settings
        settings = self.settings.copy()
        if engine_settings is None:
            engine_settings = Settings()
            engine_settings.ForceField.Type = "UFF"
        engine_name = list(engine_settings.keys())[0]
        settings.input[engine_name] = engine_settings[engine_name]

        # Lift out geometry optimization settings (now only 'Type')
        kwargs = {}
        kwargs["quiet"] = False

        for key in self.go_settings.input.ams.keys():
            if key.lower() == "geometryoptimization":
                go_sett = self.go_settings.input.ams[key]
                if "convergence" in go_sett:
                    convkeys = [k.lower() for k in go_sett.convergence.keys()]
                    if "Energy" in convkeys:
                        kwargs["convenergy"] = go_sett.Convergence.Energy
                    if "Gradients" in convkeys:
                        kwargs["convgradients"] = go_sett.Convergence.Gradients
                if "method" in go_sett:
                    kwargs["method"] = go_sett.method
                if "maxiterations" in go_sett:
                    kwargs["maxiterations"] = go_sett.maxiterations
                if "coordinatetype" in go_sett:
                    kwargs["coordinatetype"] = go_sett.coordinatetype
                if "pretendconverged" in go_sett:
                    kwargs["pretendconverged"] = go_sett.pretendconverged
        # Run the tasks
        molecule_list = []
        for i, mol in enumerate(molecules):
            molecule_list.append(("%s%i" % (name, i), mol.copy(), kwargs))
        # with AMSWorkerPool(settings, jobrunner=self.jobrunner) as job_pool :
        # maxjobs = self.jobrunner.maxjobs if self.jobrunner.parallel else 0
        # job_pool = AMSWorkerPool(settings, jobrunner=self.jobrunner)
        from scm.plams import config

        if not "default_jobmanager" in config:
            raise PlamsError("PLAMS init() not called")
        jobmanager = config.default_jobmanager
        settings.runscript.nproc = int(self.jobrunner.maxjobs / min(self.jobrunner.maxjobs, len(molecules)))
        nproc = min(self.jobrunner.maxjobs, len(molecules))
        job_pool = AMSWorkerPool(settings, nproc, workerdir_root=jobmanager.workdir)
        keepdir = False
        try:
            if task == "GeometryOptimization":
                resultlist = job_pool.GeometryOptimizations(molecule_list, watch=self.watch)
            elif task == "SinglePoint":
                resultlist = job_pool.SinglePoints(molecule_list, watch=self.watch)
            else:
                raise PlamsError("Task not implemented: %s" % (task))
        except Exception:
            keepdir = True
        finally:
            for worker in job_pool.workers:
                out, err = worker.stop(keepdir)
                # I am doing this because multiprocessing (use by ACE) also uses tempfile, and cleans the whole tempdir at the end
                worker._finalize.__call__()

        # Open the error file
        errfile = open(os.path.join(jobmanager.workdir, "amsworkerpool.err"), "a")

        # Read the results
        optimized_geometries = []
        energies = []
        for i, r in enumerate(resultlist):
            msg = "Worker failed"
            jobname = molecule_list[i][0]
            if r is not None:
                msg = r.get_errormsg()
                jobname = r.name
            if msg is not None:
                # print ('Error: ',i,msg)
                errfile.write("%-10s: %s\n" % (jobname, msg))
                optimized_geometries.append(None)
                energies.append(None)
                continue
            energy = r.get_energy()  # * Units.conversion_ratio('Hartree','kcal/mol')
            coords = r.get_main_molecule().as_array()
            optimized_geometries.append(coords)
            energies.append(energy)

        errfile.close()

        # Place the new coordinates into the molecules
        for i, (crd, en) in enumerate(zip(optimized_geometries, energies)):
            if crd is not None:
                molecules[i].from_array(crd)
            molecules[i].properties.energy = en

        return molecules, energies, resultlist

    def get_go_settings(self):
        """
        Get the settings for a geometry optimization job
        """
        # Set the settings object
        settings = self.settings.copy()
        settings.input.ams.Task = "GeometryOptimization"
        settings.update(self.go_settings)
        # settings.input.ams.GeometryOptimization.Method = self.method
        # settings.input.ams.GeometryOptimization.CoordinateType = self.coordinate_type
        return settings

    def get_sp_settings(self):
        """
        Get the settings for a geometry optimization job
        """
        # Set the settings object
        settings = self.settings.copy()
        settings.input.ams.Task = "SinglePoint"
        return settings

    def _read_results_go(self, resultlist):
        """
        Read the geometries and energies from the results objects
        """
        optimized_geometries = []
        energies = []
        for i, r in enumerate(resultlist):
            try:
                # I do it like this to make sure we hit the guardian, I think
                nsteps = r.readrkf("History", "nEntries")
            except FileError:
                pass
            # If the geometry was copied from another directory, then it is not new so we do not need it.
            if r.job.status == "failed" or not os.path.isfile(os.path.join(r.job.path, "ams.rkf")):
                optimized_geometries.append(None)
                energies.append(None)
                continue
            if nsteps is None:
                # Apparently it can still go wrong!
                nsteps = r.readrkf("History", "nEntries")
            if nsteps > 0:
                energy = r.get_property_at_step(nsteps, "Energy")  # * Units.conversion_ratio('Hartree','kcal/mol')
            else:
                try:
                    energy = r.get_energy()  # * Units.conversion_ratio('Hartree','kcal/mol')
                except FileError:
                    energy = None

            coords = r.get_main_molecule().as_array()
            optimized_geometries.append(coords)
            energies.append(energy)
            # The first time, set some variables
            if self.path is None:
                self.path = os.path.dirname(r.job.path)
                self.nsteps = nsteps
        return optimized_geometries, energies

    def _read_results_sp(self, resultlist):
        """
        Read the geometries and energies from the results objects
        """
        energies = []
        for i, r in enumerate(resultlist):
            try:
                energy = r.get_energy()  # * Units.conversion_ratio('Hartree','kcal/mol')
            except FileError:
                pass
            if r.job.status == "failed" or not os.path.isfile(os.path.join(r.job.path, "ams.rkf")):
                energies.append(None)
                continue
            energy = r.get_energy()  # * Units.conversion_ratio('Hartree','kcal/mol')
            energies.append(energy)
        return [None for en in energies], energies


def progress_monitor(results, num_jobs, event, t):
    """
    Provides regular updates on the progress of geometry optimizations in a different thread
    """
    num_done = 0
    starttime = time.time()
    width = len(str(num_jobs))
    while True:
        if event.wait(timeout=t):
            return
        try:
            num_done = len(
                [True for r in results if r.job.status not in ["created", "started", "registered", "running"]]
            )
        except RuntimeError:
            pass
        percent_done = 100.0 * num_done / num_jobs
        if percent_done > 5.0:
            dt = time.time() - starttime
            dtrem = dt / percent_done * (100 - percent_done)
            dtrem = datetime.timedelta(seconds=round(dtrem))
            trem = f", {dtrem}s remaining"
        else:
            trem = ""
        log(f"{str(num_done).rjust(width)} / {num_jobs} jobs finished:{percent_done:5.1f}%{trem}")



from collections import OrderedDict, defaultdict
from typing import List

import numpy as np
from scm.plams.core.basejob import MultiJob
from scm.plams.core.results import Results
from scm.plams.interfaces.adfsuite.amsanalysis import AMSAnalysisJob, AMSAnalysisResults
from scm.plams.tools.units import Units

__all__ = ["AMSRDFJob", "AMSMSDJob", "AMSMSDResults", "AMSVACFJob", "AMSVACFResults"]


class AMSConvenientAnalysisJob(AMSAnalysisJob):
    def __init__(self, previous_job, atom_indices=None, **kwargs):  # needs to be finished
        """
        previous_job: AMSJob
            An AMSJob with an MD trajectory. Note that the trajectory should have been equilibrated before it starts.

        All other settings can be set as for AMS

        """
        AMSAnalysisJob.__init__(self, **kwargs)

        self.previous_job = previous_job
        self.atom_indices = atom_indices

    def _get_max_dt_frames(self, max_correlation_time_fs):
        if max_correlation_time_fs is None:
            return None

        historylength = self.previous_job.results.readrkf("History", "nEntries")
        max_dt_frames = int(max_correlation_time_fs / self.previous_job.results.get_time_step())
        max_dt_frames = min(max_dt_frames, historylength // 2)
        return max_dt_frames

    def _parent_prerun(self, section):

        # use previously run previous_job
        assert self.previous_job.status != "created", "You can only pass in a finished AMSJob"

        self.settings.input.TrajectoryInfo.Trajectory.KFFileName = self.previous_job.results.rkfpath()
        if self.atom_indices and self._parent_write_atoms:
            self.settings.input[section].Atoms.Atom = self.atom_indices


class AMSMSDResults(AMSAnalysisResults):
    """Results class for AMSMSDJob"""

    def get_msd(self):
        """
        returns time [fs], msd [ang^2]
        """
        msd_xy = self.get_xy()
        time = np.array(msd_xy.x[0])  # fs
        y = np.array(msd_xy.y) * Units.convert(1.0, "bohr", "angstrom") ** 2

        return time, y

    def get_linear_fit(self, start_time_fit_fs=None, end_time_fit_fs=None):
        """
        Fits the MSD between start_time_fit_fs and end_time_fit_fs

        Returns a 3-tuple LinRegress.result, fit_x (fs), fit_y (ang^2)

        result.slope is given in ang^2/fs

        """
        from scipy.stats import linregress

        time, y = self.get_msd()
        end_time_fit_fs = end_time_fit_fs or max(time)
        start_time_fit_fs = start_time_fit_fs or self.job.start_time_fit_fs

        if start_time_fit_fs >= end_time_fit_fs:
            start_time_fit_fs = end_time_fit_fs / 2

        y = y[time >= start_time_fit_fs]
        time = time[time >= start_time_fit_fs]
        y = y[time <= end_time_fit_fs]
        time = time[time <= end_time_fit_fs]

        result = linregress(time, y)
        fit_x = time
        fit_y = result.slope * fit_x + result.intercept

        return result, fit_x, fit_y

    def get_diffusion_coefficient(self, start_time_fit_fs=None, end_time_fit_fs=None):
        """
        Returns D in m^2/s
        """
        result, _, _ = self.get_linear_fit(start_time_fit_fs=start_time_fit_fs, end_time_fit_fs=end_time_fit_fs)
        D = result.slope * 1e-20 / (6 * 1e-15)  # convert from ang^2/fs to m^2/s, divide by 6 because 3-dimensional (2d)
        return D


class AMSMSDJob(AMSConvenientAnalysisJob):
    """A convenient class wrapping around the trajectory analysis MSD tool"""

    _result_type = AMSMSDResults
    _parent_write_atoms = True

    def __init__(
        self,
        previous_job,  # needs to be finished
        max_correlation_time_fs: float = 10000,
        start_time_fit_fs: float = 2000,
        atom_indices: List[int] = None,
        **kwargs,
    ):
        """
        previous_job: AMSJob
            An AMSJob with an MD trajectory. Note that the trajectory should have been equilibrated before it starts.

        max_correlation_time_fs: float
            Maximum correlation time in femtoseconds

        start_time_fit_fs : float
            Smallest correlation time for the linear fit

        atom_indices : List[int]
            If None, use all atoms. Otherwise use the specified atom indices (starting with 1)

        kwargs: dict
            Other options to AMSAnalysisJob

        """
        AMSConvenientAnalysisJob.__init__(self, previous_job=previous_job, atom_indices=atom_indices, **kwargs)

        self.max_correlation_time_fs = max_correlation_time_fs
        self.start_time_fit_fs = start_time_fit_fs

    def prerun(self):  # noqa F811
        """
        Constructs the final settings
        """
        self._parent_prerun("MeanSquareDisplacement")  # trajectory and atom_indices handled
        max_dt_frames = self._get_max_dt_frames(self.max_correlation_time_fs)
        self.settings.input.Task = "MeanSquareDisplacement"
        self.settings.input.MeanSquareDisplacement.Property = "DiffusionCoefficient"
        self.settings.input.MeanSquareDisplacement.StartTimeSlope = self.start_time_fit_fs
        self.settings.input.MeanSquareDisplacement.MaxFrame = max_dt_frames

    def postrun(self):
        """
        Creates msd.txt, fit_msd.txt, and D.txt
        """
        time, msd = self.results.get_msd()
        with open(self.path + "/msd.txt", "w") as f:
            f.write("#Time(fs) MSD(ang^2)")
            for x, y in zip(time, msd):
                f.write(f"{x} {y}\n")

        _, fit_x, fit_y = self.results.get_linear_fit()
        with open(self.path + "/fit_msd.txt", "w") as f:
            f.write("#Time(fs), LinearFitToMSD(ang^2)")
            for x, y in zip(fit_x, fit_y):
                f.write(f"{x} {y}\n")

        D = self.results.get_diffusion_coefficient()
        with open(self.path + "/D.txt", "w") as f:
            f.write(f"{D}\n")


class AMSVACFResults(AMSAnalysisResults):
    """Results class for AMSVACFJob"""

    def get_vacf(self):
        xy = self.get_xy()
        time = np.array(xy.x[0])  # fs
        y = np.array(xy.y)

        return time, y

    def get_power_spectrum(self, max_freq=None):
        max_freq = max_freq or self.job.max_freq or 5000
        xy = self.get_xy("Spectrum")
        freq = np.array(xy.x[0])
        y = np.array(xy.y)

        y = y[freq < max_freq]
        freq = freq[freq < max_freq]

        return freq, y

    def get_acf(self):
        return self.get_vacf()

    def get_spectrum(self, max_freq=None):
        return self.get_power_spectrum(max_freq=max_freq)


class AMSVACFJob(AMSConvenientAnalysisJob):
    """A class for calculating the velocity autocorrelation function and its power spectrum"""

    _result_type = AMSVACFResults
    _parent_write_atoms = True

    def __init__(
        self,
        previous_job,  # needs to be finished
        max_correlation_time_fs=5000,  # fs
        max_freq=5000,  # cm^-1
        atom_indices=None,
        **kwargs,
    ):
        """
        previous_job: AMSJob
            An AMSJob with an MD trajectory. Note that the trajectory should have been equilibrated before it starts.

        max_correlation_time_fs: float
            Maximum correlation time in femtoseconds

        max_freq: float
            The maximum frequency for the power spectrum in cm^-1

        atom_indices: List[int]
            Atom indices (starting with 1). If None, use all atoms.

        """
        AMSConvenientAnalysisJob.__init__(self, previous_job=previous_job, atom_indices=atom_indices, **kwargs)

        self.max_correlation_time_fs = max_correlation_time_fs
        self.max_freq = max_freq

    def prerun(self):  # noqa F811
        """
        Creates final settings
        """
        self._parent_prerun("AutoCorrelation")  # trajectory and atom_indices handled
        max_dt_frames = self._get_max_dt_frames(self.max_correlation_time_fs)
        self.settings.input.Task = "AutoCorrelation"
        self.settings.input.AutoCorrelation.Property = "Velocities"
        self.settings.input.AutoCorrelation.MaxFrame = max_dt_frames

    def postrun(self):
        """
        Creates vacf.txt and power_spectrum.txt
        """
        try:
            time, vacf = self.results.get_vacf()
            with open(self.path + "/vacf.txt", "w") as f:
                f.write("#Time(fs) VACF")
                for x, y in zip(time, vacf):
                    f.write(f"{x} {y}\n")

            freq, intens = self.results.get_power_spectrum()
            with open(self.path + "/power_spectrum.txt", "w") as f:
                f.write("#Frequency(cm^-1) Intensity(arb.units)")
                for x, y in zip(freq, intens):
                    f.write(f"{x} {y}\n")
        except:
            pass


class AMSDipoleDerivativeACFResults(AMSAnalysisResults):
    """Results class for AMSVACFJob"""

    def get_acf(self):
        xy = self.get_xy()
        time = np.array(xy.x[0])  # fs
        y = np.array(xy.y)

        return time, y

    def get_ir_spectrum(self, max_freq=None):
        max_freq = max_freq or self.job.max_freq or 5000
        xy = self.get_xy("Spectrum")
        freq = np.array(xy.x[0])
        y = np.array(xy.y)

        y = y[freq < max_freq]
        freq = freq[freq < max_freq]

        return freq, y

    def get_spectrum(self, max_freq=None):
        return self.get_ir_spectrum(max_freq=max_freq)


class AMSDipoleDerivativeACFJob(AMSConvenientAnalysisJob):
    """A class for calculating the velocity autocorrelation function and its power spectrum"""

    _result_type = AMSDipoleDerivativeACFResults
    _parent_write_atoms = True

    def __init__(
        self,
        previous_job,  # needs to be finished
        max_correlation_time_fs=5000,  # fs
        max_freq=5000,  # cm^-1
        atom_indices=None,
        **kwargs,
    ):
        """
        previous_job: AMSJob
            An AMSJob with an MD trajectory. Note that the trajectory should have been equilibrated before it starts.

        max_correlation_time_fs: float
            Maximum correlation time in femtoseconds

        max_freq: float
            The maximum frequency for the power spectrum in cm^-1

        atom_indices: List[int]
            Atom indices (starting with 1). If None, use all atoms.

        """
        AMSConvenientAnalysisJob.__init__(self, previous_job=previous_job, atom_indices=atom_indices, **kwargs)

        self.max_correlation_time_fs = max_correlation_time_fs
        self.max_freq = max_freq

    def prerun(self):  # noqa F811
        """
        Creates final settings
        """
        self._parent_prerun("AutoCorrelation")  # trajectory and atom_indices handled
        max_dt_frames = self._get_max_dt_frames(self.max_correlation_time_fs)
        self.settings.input.Task = "AutoCorrelation"
        self.settings.input.AutoCorrelation.Property = "DipoleDerivativeFromCharges"
        self.settings.input.AutoCorrelation.MaxFrame = max_dt_frames

    def postrun(self):
        """
        Creates dipolederivativeacf.txt and ir_spectrum.txt
        """
        try:
            time, acf = self.results.get_acf()
            with open(self.path + "/dipolederivativeacf.txt", "w") as f:
                f.write("#Time(fs) DipoleDerivativeACF")
                for x, y in zip(time, acf):
                    f.write(f"{x} {y}\n")

            freq, intens = self.results.get_ir_spectrum()
            with open(self.path + "/ir_spectrum.txt", "w") as f:
                f.write("#Frequency(cm^-1) Intensity(arb.units)")
                for x, y in zip(freq, intens):
                    f.write(f"{x} {y}\n")
        except:
            pass


class AMSRDFResults(AMSAnalysisResults):
    """Results class for AMSRDFJob"""

    def get_rdf(self):
        """
        Returns a 2-tuple r, rdf.

        r: numpy array of float (angstrom)
        rdf: numpy array of float
        """
        xy = self.get_xy()
        r = np.array(xy.x[0])
        y = np.array(xy.y)

        return r, y


class AMSRDFJob(AMSConvenientAnalysisJob):
    _result_type = AMSRDFResults
    _parent_write_atoms = False

    def __init__(self, previous_job, atom_indices=None, atom_indices_to=None, rmin=0.5, rmax=6.0, rstep=0.1, **kwargs):
        """
        previous_job: AMSJob
            AMSJob with finished MD trajectory.

        atom_indices: List[int]
            Atom indices (starting with 1). If None, calculate RDF *from* all atoms.

        atom_indices_to: List[int]
            Atom indices (starting with 1). If None, calculate RDF *to* all atoms.

        rmin: float
            Minimum distance (angstrom)

        rmax: float
            Maximum distance (angstrom)

        rstep: float
            Bin width for the histogram (angstrom)
        """

        AMSConvenientAnalysisJob.__init__(self, previous_job=previous_job, atom_indices=atom_indices, **kwargs)
        self.atom_indices_to = atom_indices_to
        self.rmin = rmin
        self.rmax = rmax
        self.rstep = rstep

    def prerun(self):  # noqa F811
        """
        Creates the final settings. Do not call or override this method.
        """
        self._parent_prerun("RadialDistribution")
        self.settings.input.Task = "RadialDistribution"
        main_mol = self.previous_job.results.get_main_molecule()
        if not self.atom_indices:
            self.atom_indices = list(range(1, len(main_mol) + 1))
        if not self.atom_indices_to:
            self.atom_indices_to = list(range(1, len(main_mol) + 1))
        self.settings.input.RadialDistribution.AtomsFrom.Atom = self.atom_indices
        self.settings.input.RadialDistribution.AtomsTo.Atom = self.atom_indices_to
        self.settings.input.RadialDistribution.Range = f"{self.rmin} {self.rmax} {self.rstep}"

    def postrun(self):
        """
        Creates rdf.txt. Do not call or override this method.
        """
        try:
            r, gr = self.results.get_rdf()
            with open(self.path + "/rdf.txt", "w") as f:
                f.write("#r(angstrom) g(r)")
                for x, y in zip(r, gr):
                    f.write(f"{x} {y}\n")
        except:
            pass


class AMSConvenientAnalysisPerRegionResults(Results):
    def _getter(self, analysis_job_type, method, kwargs):
        assert (
            self.job.analysis_job_type is analysis_job_type
        ), f"{method}() can only be called for {analysis_job_type}, tried for type {self.job.analysis_job_type}"
        ret = {}
        for name, job in self.job.children.items():
            ret[name] = getattr(job.results, method)(**kwargs)
        return ret

    def get_diffusion_coefficient(self, **kwargs):
        return self._getter(AMSMSDJob, "get_diffusion_coefficient", kwargs)

    def get_msd(self, **kwargs):
        return self._getter(AMSMSDJob, "get_msd", kwargs)

    def get_linear_fit(self, **kwargs):
        return self._getter(AMSMSDJob, "get_linear_fit", kwargs)

    def get_vacf(self, **kwargs):
        return self._getter(AMSVACFJob, "get_vacf", kwargs)

    def get_power_spectrum(self, **kwargs):
        return self._getter(AMSVACFJob, "get_power_spectrum", kwargs)


class AMSConvenientAnalysisPerRegionJob(MultiJob):
    _result_type = AMSConvenientAnalysisPerRegionResults

    def __init__(self, previous_job, analysis_job_type, name=None, regions=None, per_element=False, **kwargs):
        MultiJob.__init__(self, children=OrderedDict(), name=name or "analysis_per_region")
        self.previous_job = previous_job
        self.analysis_job_type = analysis_job_type
        self.analysis_job_kwargs = kwargs
        self.regions_dict = regions
        self.per_element = per_element

    @staticmethod
    def get_regions_dict(molecule, per_element: bool = False):
        regions_dict = defaultdict(lambda: [])
        for i, at in enumerate(molecule, 1):
            regions = set([at.properties.region]) if isinstance(at.properties.region, str) else at.properties.region
            if len(regions) == 0:
                region_name = "NoRegion" if not per_element else f"NoRegion_{at.symbol}"
                regions_dict[region_name].append(i)
            for region in regions:
                region_name = region if not per_element else f"{region}_{at.symbol}"
                regions_dict[region_name].append(i)
            regions_dict["All"].append(i)
            if per_element:
                regions_dict[f"All_{at.symbol}"].append(i)

        return regions_dict

    def prerun(self):  # noqa F811
        regions_dict = self.regions_dict or self.get_regions_dict(
            self.previous_job.results.get_main_molecule(), per_element=self.per_element
        )

        for region in regions_dict:
            self.children[region] = self.analysis_job_type(
                previous_job=self.previous_job,
                name=region,
                atom_indices=regions_dict[region],
                **self.analysis_job_kwargs,
            )

    @staticmethod
    def get_mean_std_per_region(list_of_jobs, function_name, **kwargs):
        """
        list_of_jobs: List[AMSConvenientAnalysisPerRegionJob]
            List of jobs over which to average

        function_name: str
            e.g. 'get_msd', 'get_power_spectrum'

        """

        if isinstance(list_of_jobs, dict):
            list_of_jobs = [x for x in list_of_jobs.values()]

        all_x = defaultdict(lambda: [])
        all_y = defaultdict(lambda: [])
        for vacfjob in list_of_jobs:
            # let's calculate mean and std for each region
            results_dict = getattr(vacfjob.results, function_name)(**kwargs)
            for region_name, ret_value in results_dict.items():
                if np.isscalar(ret_value):
                    all_x[region_name].append(np.atleast_1d(ret_value))
                elif len(ret_value) == 2:
                    all_x[region_name].append(np.atleast_1d(ret_value[0]))
                    all_y[region_name].append(np.atleast_1d(ret_value[1]))

        mean_x = {}
        std_x = {}
        mean_y = {}
        std_y = {}
        for region_name, values in all_x.items():
            mean_x[region_name] = np.mean(values, axis=0)
            std_x[region_name] = np.std(values, axis=0)

            if len(all_y) > 0:
                mean_y[region_name] = np.mean(all_y[region_name], axis=0)
                std_y[region_name] = np.std(all_y[region_name], axis=0)

        if len(mean_y) > 0:
            return mean_x, std_x, mean_y, std_y
        else:
            return mean_x, std_x



from scm.plams import Atom, Molecule
from scm.plams.interfaces.molecule.rdkit import from_smiles, to_smiles


def test_smiles():
    # Super basic test: water molecule

    mol = Molecule()
    mol.add_atom(Atom(symbol="O", coords=(0.0, 0.0, 0.0)))
    mol.add_atom(Atom(symbol="H", coords=(1.0, 0.0, 0.0)))
    mol.add_atom(Atom(symbol="H", coords=(0.0, 1.0, 0.0)))

    assert to_smiles(mol) == "O"
    assert to_smiles(mol, short_smiles=False) == "[H]O[H]"

    # Note! in the list of smiles below I had to escape some backslashes (so '\\' was originally a '\')

    a_bunch_of_smiles = [
        "C(=O)(O)C(C)(C)Oc1ccc(cc1)CC(=O)Nc1cc(cc(c1)C)C",
        "CC(C)([C@@H](C(=O)NC)NC(=O)[C@@H](CN(C=O)O)CC(C)C)C",
        "O=C1N2[C@@H](C(=O)[O-])C(C)(C)[S@](=O)[C@@H]2[C@@H]1NC(=O)Cc1ccccc1",
        "c1cc(cc2c1CC[NH2+]C2)S(=O)(=O)N",
        "OC[C@H]1O[C@H]([C@@H]([C@@H]1O)O)n1cnc2c(N)nccc12",
        "C1Nc2nc(N)[nH]c(=O)c2N=C1CO",
        "n1c(C)[nH]c(=O)c2cc(ccc12)CN(c1sc(cc1)C(=O)N[C@@H](CCC(=O)O)C(=O)O)C",
        "c12c(cccc1)n(c(c2c1ccc(cc1)F)/C=C/[C@H](C[C@@H](O)CC(=O)O)O)C(C)C",
        "c12c(c(nc([nH+]1)N)N)c(ccc2)Sc1ccccc1",
        "Cc1nc(N)c(C[n+]2csc(CCO)c2C)cn1",
        "n1c([nH+]c(c(c1N)c1ccc(cc1)Cl)CC)N",
        "c1(nnc(NC(=O)C)s1)S(=O)(=O)[NH-]",
        "[C@H](C(=O)O)(Cc1ccccc1)[C@@H](C(=O)O)Cc1cc2OCOc2cc1",
        "c1(ccccc1)Cc1n(c(=O)[nH]c(=O)c1C(C)C)COCc1ccccc1",
        "O=C(O)C[C@@H](C(=O)O)NC(=O)Cc1c[nH]c2c1cccc2",
        "N1C(=O)/C(=C\\Nc2ccc(cc2)S(=O)(=O)NC)/c2ccccc12",
        "c1(O)cccc(c1C)C(=O)N[C@@H](Cc1ccccc1)[C@H](O)C(=O)N1CSC(C)(C)[C@H]1C(=O)NCc1c(cccc1)C",
        "[C@H]1(NC(=[NH2+])N)[C@H]([C@H](C(CC)CC)NC(=O)C)[C@H](O)[C@@H](C(=O)O)C1",
        "c1cc2c(c(c1)C)cc(n2Cc1cccc(c1)C(=[NH2+])N)C(=O)NCc1cc(cc(c1)Cl)Cl",
        "c12c(cccc1cccc2)CC(=O)O",
        "C1=CC(=O)C=C2CC[C@@H]3[C@@]([C@@]12C)([C@H](C[C@]1([C@H]3C[C@H]([C@@]1(C(=O)CO)O)C)C)O)F",
        "C1(=O)c2c(CO1)c(C)c(c(c2O)C/C=C(/CCC(=O)O)\\C)OC",
        "[NH3+][C@H](C(=O)O)CCCNC(=[NH2+])NCCC",
        "[C@@]1(c2cc(Oc3c(ccc(c3)[C@](c3n(cnc3)C)(C)[NH3+])C#N)ccc2)(C(=O)N(CCCC1)C)CC",
        "C1CN(CC1)C(=O)[C@H](C(C)C)[NH3+]",
        "O=C(O)[C@H](O)C(C)(C)CO",
        "[nH]1c(CCCC)nc2c(=O)[nH][nH]c(=O)c12",
        "O=c1[nH]c(=O)cnn1c1cc(C)c(c(c1)C)Oc1ccc(c(c1)C(C)C)O",
        "O=C(O)Cc1cc(c(Oc2ccc(O)c(c2)C(C)C)c(c1)Cl)Cl",
        "OC[C@@H]1[C@@H](O)C[C@]2(n3c(=O)[nH]c(=O)c(c3)C)[C@H]1C2",
        "[NH3+][C@H](C(=O)O)Cc1ccc(cc1)O",
        "OCc1cccc(Nc2ncc3cc(c(=O)n(c3n2)C)c2c(Cl)cccc2Cl)c1",
        "S(=O)(=O)(c1ccc(cc1)n1c(c2ccc(C)cc2)cc(C(F)(F)F)n1)[NH-]",
        "[NH2+]=C(N)c1ccc2cc(ccc2c1)C(=O)Nc1ccccc1",
        "O=C1N(Cc2ccc(F)cc2)C(=O)[C@H]2[C@H]3N(CCC3)[C@@H](c3ccc(cc3)C(=[NH2+])N)[C@@H]12",
        "n1cc(ccc1)[C@H]1[N@@H+](CCC1)C",
        "OC[C@H]1O[C@@H](n2ccc(nc2=O)N)C(F)(F)[C@@H]1O",
        "n1(c(nc(c1c1ccnc(n1)NC1CC1)c1cc(c(cc1)Cl)Cl)[C@H]1CC[N@@H+](CC1)C)CCC",
        "CSC[C@H]1[NH2+][C@H]([C@H](O)[C@@H]1O)c1c[nH]c2c(=O)[nH]cnc12",
        "c12/C(=N\\O)/C(=C\\3/C(=O)Nc4c3cccc4)/Nc1cccc2",
        "c1cc([C@@H](C(=O)O)C)ccc1c1ccccc1",
        "C[C@H]([NH3+])[P@](=O)(O)C[C@H](C(=O)N[C@@H](C)C(=O)O)Cc1ccc(cc1)c1ccccc1",
        "N(C(=O)[C@@H]([C@@H](C(=O)NO)O)CC(C)C)[C@@H](C(C)(C)C)C(=O)NC",
        "C(C)(C)SCC[C@@H](N)[C@H](O)C(=O)NNC(=O)c1cccc(c1)Cl",
        "c1cc(ccc1)c1ccc(cc1F)[C@H](C)C(=O)O",
        "c1([nH+]c2c(c(n1)N)C[C@@H](CC2)CN(c1cc(c(c(c1)OC)OC)OC)C)N",
        "c1(cc(cc(c1)/C=C/c1ccc(cc1)O)O)O",
        "[C@H]1([C@@H](Oc2ccc(O)cc2S1)c1ccc(cc1)OCC[NH+]1CCCCC1)c1ccc(O)cc1",
        "OCC(C)(C)[C@@H](O)C(=O)NCCC(=O)O",
        "c1c(c(ccc1F)C(=O)NCc1nc2c(s1)c(F)cc(F)c2F)OCC(=O)O",
        "c1cc(cnc1)c1ccnc(n1)Nc1c(ccc(c1)NC(=O)c1ccc(cc1)CN1CC[N@H+](C)CC1)C",
        "COc1nc(C)nc(N/C(=N/S(=O)(=O)c2ccccc2Cl)/O)n1",
        "O=C(O)CCCn1c2ccccc2c2c1cccc2",
        "[NH2+]1C[C@@H]([C@@H]([C@H]1C(=O)O)CC(=O)O)C(=C)C",
        "CC/C(=C(\\c1ccc(O)cc1)/CC)/c1ccc(O)cc1",
        "OCCOCn1c(=O)[nH]c(=O)c(c1)Cc1ccccc1",
        "O=C1NCC/C(=C\\2/NC(=NC2=O)N)/c2c1[nH]cc2",
        "c1n(cnc1C(=O)N)[C@@H](CO)CCn1ccc2c1cc(cc2)NC(=O)CCc1ccccc1",
        "OC[C@@H](CC)Nc1nc2n(C(C)C)cnc2c(n1)NCc1ccccc1",
        "Clc1c(CN2CCCC2=[NH2+])[nH]c(=O)[nH]c1=O",
        "n1c(nc2n(C(C)C)cnc2c1Nc1ccc(c(Cl)c1)C(=O)O)N[C@H](C(C)C)CO",
        "n1(cnc2c(=O)[nH]c(N)nc12)CCCCC(F)(F)P(=O)(O)O",
        "s1ccnc1NC(=O)c1c(cc(c(c1)Sc1n(ccn1)C)F)N",
        "c1c(c(cc(c1)C(=O)O)NC(CC)CC)N1[C@@](CCC1=O)(CO)C[NH3+]",
        "O=C1[C@@H]2CCCN2C(=O)CN1",
        "OC[C@H]1O[C@H](C[C@@H]1O)n1c(=O)[nH]c(=O)c(C)c1",
        "O=C1O[C@](CN1)(C)c1ccc(OC)c(c1)OCCC",
        "O=C(c1ccc(OC(F)F)c(OCC2CC2)c1)Nc1c(Cl)cncc1Cl",
        "c1ccc2c3c([nH]c2c1)[C@H](N1C(=O)CN(C(=O)[C@H]1C3)C)c1ccc2OCOc2c1",
        "O=S(=O)(NCC1CC1)c1ccc(c(c1)Nc1oc(cn1)c1cc(ccc1)c1c[nH+]ccc1)OC",
        "Oc1cc(N[C@@H](C(=O)NS(=O)(=O)c2cccc(N)c2)c2c(F)c(OCC)cc(OCC)c2)ccc1C(=[NH2+])N",
        "c1(cc(c(cc1)F)C)S(=O)(=O)N[C@@H](C(=O)NO)C1CCOCC1",
        "C(=C(\\NC(=O)c1ccccc1)/C(=O)O)\\c1ccc(cc1)Oc1c(cccc1)Br",
        "c1c(ccc(c1)F)c1c(c2nc(N[C@H](c3ccccc3)C)ncc2)n(n(C2CC[NH2+]CC2)c1=O)C",
        "c1cc(F)ccc1S(=O)(=O)C[C@@](O)(C)C(=O)Nc1cc(c(cc1)C#N)C(F)(F)F",
        "c1(cc(ccc1)C[NH3+])[C@H]1CCN(C(=O)c2cc(CCc3ccccc3)cnc2)CC1",
        "COc1ccc(cc1)c1c2c(NCCO)ncnc2oc1c1ccc(OC)cc1",
        "COc1ccc(cc1)c1c(C(=O)NCC)n[nH]c1c1cc(Cl)c(O)cc1O",
        "C(=O)(c1ccccc1)NNc1n[nH]c(=S)n1N",
        "Brc1n(c(C(C)C)c(n1)NC(=O)C)C",
        "c1(c(n(cc1NS(=O)(=O)c1ccccc1)C(C)(C)C)N(C)C)c1ccccc1",
        "Brc1ccc(cc1)/C(=C/c1ccccc1)/C1=CC(=O)N(C1)C(=O)C",
        "c12c(ccc(c1c(=O)cc(C(=O)[O-])o2)OCCC(C)C)CC=C",
        "S(=O)(=O)(c1ccc(N(=O)=O)cc1)Cc1ccc(N(C)C)cc1",
        "s1cc(n(/c/1=C/N(=O)=O)CCNC(=O)C)c1ccccc1",
        "c1cc(c2c(c3ccccc3)c(N(C(=O)C)C(=O)C)on2)ccc1",
        "c1(cc(c(cc1)NC(=O)C)SC)c1ccc(cc1)NC(=O)C",
        "n1cnc2c(c1N)ncn2CCC(=O)NCCc1c[nH]c2c1cccc2",
        "ClCS(=O)(=O)/C(=C(\\c1ccccc1)/N1CCOCC1)/C",
        "n1c(Cl)c(ccc1)NCc1nc(Cl)c(N)cc1",
        "c12c(=O)n(c3ccccc3)[nH]c1nc(Nc1ccccc1)s2",
        "s1c(nc(c1N(c1ccccc1)C)C)Nc1ccccc1",
        "s1c(nc(c1N(c1ccccc1)C)c1ccccc1)Nc1ccccc1",
        "c1cc(ccc1C(=C1CCCCC1)c1ccc(cc1)OC(=O)C)OC(=O)C",
        "n1cnc2c(c1N)ncn2CCC(=O)NCCc1ccc(cc1)O",
        "N(C)(C)c1ccc(cc1)C(=[NH2+])c1ccc(N(C)C)cc1",
        "c1(CCC(=O)O)cc(c2cccccc12)CCC(=O)O",
        "O1CC[NH+](CC1)CCNc1ccc(n[nH+]1)c1ccccc1",
        "n1c(nc(cc1)N)SSc1nccc(n1)N",
        "Fc1c(=O)[nH]c(=O)n(c1)c1nc(nc(c1)C)Cc1ccc(OC)cc1",
    ]

    # level = 2 means 'build a molecule label using atoms connectivity and bond orders'
    level = 2

    for smiles in a_bunch_of_smiles:
        mol = from_smiles(smiles)
        my_smiles = to_smiles(mol)
        my_mol = from_smiles(my_smiles)
        assert my_mol.label(level) == mol.label(level)



#!/usr/bin/env amspython
import os

from scm.plams import *


def main():
    # the first argument needs to be a file readable by ASE
    original_trajectory = os.path.expandvars(
        "$AMSHOME/scripting/scm/params/tests/fixtures/results_importer/vasp/OUTCAR"
    )
    convert_to_ams_rkf_with_bond_guessing(original_trajectory, "converted-to-ams.rkf")
    # convert_to_ams_rkf_with_bond_guessing('somefile.xyz', 'converted-to-ams.rkf')


def convert_to_ams_rkf_with_bond_guessing(filename, outfile="out.rkf", task="moleculardynamics", timestep=0.5):
    temp_traj = "out.traj"
    file_to_traj(filename, temp_traj)
    traj_to_rkf(temp_traj, outfile, task=task, timestep=timestep)

    init()
    # config.log.stdout = 0
    # config.erase_workdir = True   # to remove workdir, only use this if you're not already inside another PLAMS workflow

    s = Settings()
    s.input.ams.task = "replay"
    s.input.lennardjones
    s.input.ams.replay.file = os.path.abspath(outfile)
    s.input.ams.properties.molecules = "yes"
    s.input.ams.properties.bondorders = "yes"
    s.runscript.nproc = 1
    job = AMSJob(settings=s, name="rep")
    job.run()
    job.results.wait()
    cpkf = os.path.expandvars("$AMSBIN/cpkf")
    os.system(f'sh "{cpkf}" "{job.results.rkfpath()}" "{outfile}" History Molecules')
    delete_job(job)
    finish()

    os.remove(temp_traj)


if __name__ == "__main__":
    main()



import warnings

import numpy as np

from scm.plams.core.errors import ResultsError
from scm.plams.core.functions import config, log
from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.scmjob import SCMJob, SCMResults
from scm.plams.tools.periodic_table import PT
from scm.plams.tools.units import Units

__all__ = ["ADFJob", "ADFResults"]


class ADFResults(SCMResults):
    """A specialized |SCMResults| subclass for accessing the results of |ADFJob|."""

    _kfext = ".t21"
    _rename_map = {"TAPE{}".format(i): "$JN.t{}".format(i) for i in range(10, 100)}

    def get_properties(self):
        """get_properties()
        Return a dictionary with all the entries from ``Properties`` section in the main KF file (``$JN.t21``).
        """
        ret = {}
        for sec, var in self._kf:
            if sec == "Properties":
                ret[var] = self.readkf(sec, var)
        return ret

    def get_main_molecule(self):
        """get_main_molecule()
        Return a |Molecule| instance based on the ``Geometry`` section in the main KF file (``$JN.t21``).

        For runs with multiple geometries (geometry optimization, transition state search, intrinsic reaction coordinate) this is the **final** geometry. In such a case, to access the initial (or any intermediate) coordinates please use :meth:`get_input_molecule` or extract coordinates from section ``History``, variables ``xyz 1``, ``xyz 2`` and so on. Mind the fact that all coordinates written by ADF to ``History`` section are in bohr and internal atom order::

            mol = results.get_molecule(section='History', variable='xyz 1', unit='bohr', internal=True)
        """
        return self.get_molecule(section="Geometry", variable="xyz InputOrder", unit="bohr")

    def get_input_molecule(self):
        """get_input_molecule()
        Return a |Molecule| instance with initial coordinates.

        All data used by this method is taken from ``$JN.t21`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        if ("History", "nr of geometries") in self._kf:
            return self.get_molecule(section="History", variable="xyz 1", unit="bohr", internal=True)
        return self.get_main_molecule()

    def get_energy(self, unit="au"):
        """get_energy(unit='au')
        Return final bond energy, expressed in *unit*.
        """
        return self._get_single_value("Energy", "Bond Energy", unit)

    def get_dipole_vector(self, unit="au"):
        """get_dipole_vector(unit='au')
        Return the dipole vector, expressed in *unit*.
        """
        prop = self.get_properties()
        if "Dipole" in prop:
            return Units.convert(prop["Dipole"], "au", unit)
        raise ResultsError("'Dipole' not present in 'Properties' section of {}".format(self._kfpath()))

    def get_gradients(self, energy_unit="au", dist_unit="bohr", eUnit=None, lUnit=None):
        """get_gradients(energy_unit='au', dist_unit='bohr')
        Return the cartesian gradients from the 'Gradients_InputOrder' field of the 'GeoOpt' Section in the kf-file, expressed in given units. Returned value is a numpy array with shape (nAtoms,3).
        """
        if eUnit:
            log(
                "Deprecated Keyword eUnit used in ADFResults.get_gradients, update your script! Overwriting energy_unit with the given argument.",
                1,
            )
            warnings.warn("eUnit is deprecated, use energy_unit instead.", category=DeprecationWarning)
            energy_unit = eUnit
        if lUnit:
            log(
                "Deprecated Keyword lUnit used in ADFResults.get_gradients, update your script! Overwriting energy_unit with the given argument.",
                1,
            )
            warnings.warn("lUnit is deprecated, use dist_unit instead.", category=DeprecationWarning)
            dist_unit = lUnit
        gradients = np.array(self.readkf("GeoOpt", "Gradients_InputOrder"))
        gradients.shape = (-1, 3)
        gradients *= Units.conversion_ratio("au", energy_unit) / Units.conversion_ratio("bohr", dist_unit)
        return gradients

    def _extract_hessian(self, section, variable, internal_order):
        """_extract_hessian(section, variable, internal_order)
        Extract Hessian from *section*/*variable* of the TAPE21 file. Reorder from internal to input order, if *internal_order* is ``True``.
        """
        hess_int = np.array(self.readkf(section, variable))
        n = int((len(hess_int) / 9 + 1) ** 0.5)
        hess_int.shape = (3 * n, 3 * n)
        if internal_order:
            hess_inp = np.zeros(hess_int.shape)
            mapping = self._int2inp()
            for i in range(n):
                for j in range(n):
                    ii, jj = mapping[i] - 1, mapping[j] - 1
                    hess_inp[3 * i : 3 * i + 3, 3 * j : 3 * j + 3] = hess_int[3 * ii : 3 * ii + 3, 3 * jj : 3 * jj + 3]
            return hess_inp
        return hess_int

    def get_hessian(self):
        """get_hessian()
        Try extracting Hessian, either analytical or numerical, whichever is present in the TAPE21 file, in the input order. Returned value is a square numpy array of size 3*nAtoms.
        """
        if ("Hessian", "Analytical Hessian") in self._kf:
            return self._extract_hessian("Hessian", "Analytical Hessian", True)
        if ("Freq", "Hessian_complete") in self._kf:
            return self._extract_hessian("Freq", "Hessian_complete", True)
        raise ResultsError("auto_hessian: Hessian does not seem to be present in t21 file.")

    def get_energy_decomposition(self, unit="au"):
        """get_energy(unit='au')
        Return a dictionary with energy decomposition terms, expressed in *unit*.

        The following keys are present in the returned dictionary: ``Electrostatic``, ``Kinetic``, ``Coulomb``, ``XC``. The sum of all the values is equal to the value returned by :meth:`get_energy`.
        Note that additional contributions might be included, those are up to now: ``Dispersion``.
        """
        ret = {}
        ret["Electrostatic"] = self._get_single_value("Energy", "Electrostatic Energy", unit)
        ret["Kinetic"] = self._get_single_value("Energy", "Kinetic Energy", unit)
        ret["Coulomb"] = self._get_single_value("Energy", "Elstat Interaction", unit)
        ret["XC"] = self._get_single_value("Energy", "XC Energy", unit)
        if ("Energy", "Dispersion Energy") in self._kf:
            ret["Dispersion"] = self._get_single_value("Energy", "Dispersion Energy", unit)
        return ret

    def get_frequencies(self, unit="cm^-1"):
        """get_frequencies(unit='cm^-1')
        Return a numpy array of vibrational frequencies, expressed in *unit*.
        """
        freqs = np.array(self.readkf("Freq", "Frequencies"))
        return freqs * Units.conversion_ratio("cm^-1", unit)

    def get_timings(self):
        """get_timings()

        Return a dictionary with timing statistics of the job execution. Returned dictionary contains keys ``cpu``, ``system`` and ``elapsed``. The values are corresponding timings, expressed in seconds.
        """
        last = self.grep_output(" Total Used : ")[-1].split()
        ret = {}
        ret["elapsed"] = float(last[-1])
        ret["system"] = float(last[-3])
        ret["cpu"] = float(last[-5])
        return ret

    def _atomic_numbers_input_order(self):
        """_atomic_numbers_input_order()
        Return a list of atomic numbers, in the input order.
        """
        n = self.readkf("Geometry", "nr of atoms")
        tmp = self.readkf("Geometry", "atomtype").split()
        atomtypes = {i + 1: PT.get_atomic_number(tmp[i]) for i in range(len(tmp))}
        atomtype_idx = self.readkf("Geometry", "fragment and atomtype index")[-n:]
        atnums = [atomtypes[i] for i in atomtype_idx]
        return self.to_input_order(atnums)

    def _int2inp(self):
        """_int2inp()
        Get mapping from the internal atom order to the input atom order.
        """
        aoi = self.readkf("Geometry", "atom order index")
        n = len(aoi) // 2
        return aoi[:n]

    def recreate_molecule(self):
        """Recreate the input molecule for the corresponding job based on files present in the job folder. This method is used by |load_external|."""
        if self._kfpresent():
            return self.get_input_molecule()
        return None

    def recreate_settings(self):
        """Recreate the input |Settings| instance for the corresponding job based on files present in the job folder. This method is used by |load_external|."""
        if self._kfpresent():
            if ("General", "Input") in self._kf:
                tmp = self.readkf("General", "Input")
                user_input = "\n".join([tmp[i : 160 + i].rstrip() for i in range(0, len(tmp), 160)])
            else:
                user_input = self.readkf("General", "user input")
            try:
                from scm.plams.interfaces.adfsuite.inputparser import InputParserFacade

                inp = InputParserFacade().to_settings("adf", user_input)
            except:
                log("Failed to recreate input settings from {}".format(self._kf.path), 5)
                return None

            s = Settings()
            s.input = inp
            del s.input.atoms
            s.soft_update(config.job)
            return s
        return None


class ADFJob(SCMJob):
    _result_type = ADFResults
    _command = "adf"

    def _serialize_mol(self):
        for i, atom in enumerate(self.molecule):
            smb = self._atom_symbol(atom)
            suffix = ""
            if "adf" in atom.properties and "fragment" in atom.properties.adf:
                suffix += "f={fragment} "
            if "adf" in atom.properties and "block" in atom.properties.adf:
                suffix += "b={block}"

            self.settings.input.atoms["_" + str(i + 1)] = ("{:>5}".format(i + 1)) + atom.str(
                symbol=smb, suffix=suffix, suffix_dict=atom.properties.adf
            )

    def _remove_mol(self):
        if "atoms" in self.settings.input:
            del self.settings.input.atoms



#!/usr/bin/env plams
import numpy as np

# Calcualte bond energy of He dimers for a series of bond
# distances using ADF

# type of atoms
atom1 = "He"
atom2 = "He"

# interatomic distance values
dmin = 2.2
dmax = 4.2
step = 0.2

# create a list with interatomic distances
distances = np.arange(dmin, dmax, step)

# calculation parameters (single point, TZP/PBE+GrimmeD3)
sett = Settings()
sett.input.ams.task = "SinglePoint"
sett.input.adf.basis.type = "TZP"
sett.input.adf.xc.gga = "PBE"
sett.input.adf.xc.dispersion = "Grimme3"

energies = []
for d in distances:
    mol = Molecule()
    mol.add_atom(Atom(symbol=atom1, coords=(0.0, 0.0, 0.0)))
    mol.add_atom(Atom(symbol=atom2, coords=(d, 0.0, 0.0)))
    job = AMSJob(molecule=mol, settings=sett, name=f"dist_{d:.2f}")
    job.run()
    energies.append(job.results.get_energy(unit="kcal/mol"))

# print
print("== Results ==")
print("d[A]    E[kcal/mol]")
for d, e in zip(distances, energies):
    print(f"{d:.2f}    {e:.3f}")



#!/usr/bin/env amspython
import os
import time

from ase.io import read
from scm.plams import *
from scm.plams.interfaces.adfsuite.ase_calculator import AMSCalculator
from sella import Sella

"""

This examples runs a transition state search with the Sella python package and
compares results and timings to the builtin transition state search in AMS (using a DFTB hessian to initialize the DFT transition state search).

This example requires an ADF license.

"""


def adf_settings():
    s = Settings()
    s.input.adf.basis.type = "DZP"
    s.input.adf.basis.core = "None"
    s.input.adf.xc.gga = "pbe"
    s.input.adf.symmetry = "nosym"
    return s


def run_sella(molecule):
    """Run a TS search with Sella"""
    s = adf_settings()
    s.input.ams.task = "SinglePoint"
    s.input.ams.properties.gradients = "Yes"

    atoms = toASE(molecule)

    if os.path.exists("sella.log"):
        os.remove("sella.log")
    if os.path.exists("sella.traj"):
        os.remove("sella.traj")

    with AMSCalculator(settings=s, amsworker=True) as calc:
        atoms.calc = calc
        opt = Sella(atoms, internal=True, logfile="sella.log", trajectory="sella.traj")
        opt.run(fmax=0.027, steps=50)  # 0.027 eV/ang roughly matches AMS default criterion 0.001 Ha/ang

    traj = read("sella.traj", ":")
    print(f"Sella converged after {len(traj)} single-point calculations")
    print(f"Final energy: {atoms.get_potential_energy()} eV")

    optimized_mol = fromASE(atoms)
    return optimized_mol


def run_ams(molecule):
    """Run DFT transition state search but use initial hessian calculated with DFTB"""
    s = adf_settings()
    s.input.ams.task = "TransitionStateSearch"
    s.input.ams.geometryoptimization.initialhessian.type = "CalculateWithFastEngine"
    job = AMSJob(settings=s, molecule=molecule, name="ams_ts")
    job.run()
    print(f"AMS finished after {len(job.results.get_history_property('Energy'))} iterations")
    print(f"AMS optimized energy: {job.results.get_energy(unit='eV')} eV")


def get_molecule():
    return AMSJob.from_input(
        """
        System
          atoms
             C         0.049484    0.042994    0.000000
             H        -0.068980    0.638928   -0.915972
             H        -0.068980    0.638928    0.915972
             H        -0.841513   -0.626342    0.000000
             H         0.555494   -1.148227    0.000000
             Hg        2.303289   -0.007233    0.000000
             Cl        4.429752    0.776056    0.000000
             Cl        1.342057   -2.676083    0.000000
          end
        End

    """
    ).molecule[""]


def main():
    os.environ["OMP_NUM_THREADS"] = "1"
    init()
    mol = get_molecule()

    start = time.time()
    run_sella(mol)
    print(f"Sella finished in {time.time()-start} seconds")

    start = time.time()
    run_ams(mol)
    print(f"AMS finished in {time.time()-start} seconds")

    finish()


if __name__ == "__main__":
    main()



#!/usr/bin/env plams
import os

import matplotlib.pyplot as plt
import numpy as np
from scm.plams import *


def run_md():
    mol = packmol(from_smiles("O"), n_molecules=16, density=1.0)
    s = Settings()
    s.input.ams.Task = "MolecularDynamics"
    s.input.ReaxFF.ForceField = "Water2017.ff"
    s.input.ams.MolecularDynamics.CalcPressure = "Yes"
    s.input.ams.MolecularDynamics.InitialVelocities.Temperature = 300
    s.input.ams.MolecularDynamics.Trajectory.SamplingFreq = 1
    s.input.ams.MolecularDynamics.TimeStep = 0.5
    s.input.ams.MolecularDynamics.NSteps = 2000
    s.runscript.nproc = 1
    os.environ["OMP_NUM_THREADS"] = "1"
    job = AMSJob(settings=s, molecule=mol, name="md")
    job.run()
    return job


def plot_results(results):
    plt.clf()
    times, vacf = results.get_velocity_acf(start_fs=0, max_dt_fs=250, normalize=False)
    normalized_vacf = vacf / vacf[0]
    plt.plot(times, normalized_vacf)
    plt.xlabel("Time (fs)")
    plt.ylabel("Velocity autocorrelation function")
    plt.title("Velocity autocorrelation function")
    plt.savefig("plams_vacf.png")
    A = np.stack((times, normalized_vacf), axis=1)
    np.savetxt("plams_vacf.txt", A, header="Time(fs) VACF")

    plt.clf()
    t_D, D = results.get_diffusion_coefficient_from_velocity_acf(times, vacf)
    plt.plot(t_D, D)
    plt.xlabel("Time (fs)")
    plt.ylabel("D (m²s⁻¹)")
    plt.title("Diffusion coefficient")
    plt.savefig("plams_vacf_D.png")
    A = np.stack((t_D, D), axis=1)
    np.savetxt("plams_vacf_D.txt", A, header="time(fs) D(m^2*s^-1)")

    plt.clf()
    freq, intensities = results.get_power_spectrum(times, vacf, number_of_points=1000)
    plt.plot(freq, intensities)
    plt.xlabel("Frequency (cm⁻¹)")
    plt.ylabel("Power spectrum (arbitrary units)")
    plt.title("Power spectrum")
    plt.savefig("plams_power_spectrum.png")
    A = np.stack((freq, intensities), axis=1)
    np.savetxt("plams_power_spectrum.txt", A, header="Frequency(cm^-1) PowerSpectrum")

    plt.clf()
    t, viscosity = results.get_green_kubo_viscosity(start_fs=0, max_dt_fs=250)  # do not do this for NPT simulations
    plt.plot(t, viscosity)
    plt.xlabel("Time (fs)")
    plt.ylabel("Viscosity (mPa*s)")
    plt.title("Viscosity")
    plt.savefig("plams_green_kubo_viscosity.png")
    A = np.stack((t, viscosity), axis=1)
    np.savetxt("plams_green_kubo_viscosity.txt", A, header="Time(fs) Viscosity(mPa*s)")

    plt.clf()
    z, density = results.get_density_along_axis(axis="z", density_type="mass", bin_width=0.2, atom_indices=None)
    plt.plot(z, density)
    plt.xlabel("z coordinate (Å)")
    plt.ylabel("Density (g/cm³)")
    plt.title("Density along z")
    plt.savefig("plams_density_along_z.png")
    A = np.stack((z, density), axis=1)
    np.savetxt("plams_density_along_z.txt", A, header="z(angstrom) density(g/cm^3)")


def main():
    job = run_md()
    # alternatively:
    # job = AMSJob.load_external('/path/to/ams.rkf')
    plot_results(job.results)


if __name__ == "__main__":
    main()



import shutil
from os import symlink
from os.path import basename
from os.path import join as opj
from os.path import relpath

import numpy as np

from scm.plams.core.basejob import SingleJob
from scm.plams.core.errors import JobError
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
from scm.plams.mol.molecule import Molecule
from scm.plams.tools.units import Units

__all__ = ["ORCAJob", "ORCAResults"]


class ORCAResults(Results):
    """A class for ORCA results."""

    def get_runtime(self):
        """Return runtime in seconds from output."""
        runtimeString = self.grep_output("TOTAL RUN TIME")[-1]
        runtimeList = [int(x) for x in runtimeString.split()[3::2]]
        assert len(runtimeList) == 5  # days hours minutes seconds milliseconds
        runtime = sum([t * m for t, m in zip(runtimeList, [24 * 60 * 60, 60 * 60, 60, 1, 0.001])])
        return runtime

    def get_timings(self):
        """Return timings section as dictionary. Units are seconds."""
        timingsSection = self.get_output_chunk("Timings for individual modules:", end="****")
        ret = {}
        for line in timingsSection:
            line = line.strip()
            if not line:
                continue
            name, data = line.split("...")
            ret[name.strip()] = float(data.split()[0])
        return ret

    def check(self):
        """Returns true if ORCA TERMINATED NORMALLY is in the output"""
        return bool(self.grep_output("ORCA TERMINATED NORMALLY"))

    def check_go_conv(self):
        """Returns true if THE OPTIMIZATION HAS CONVERGED is in the output"""
        return bool(self.grep_output("THE OPTIMIZATION HAS CONVERGED"))

    def check_scf_conv(self):
        """Returns true if SCF NOT CONVERGED AFTER is NOT in the output"""
        return not bool(self.grep_output("SCF NOT CONVERGED AFTER"))

    def _get_energy_correct_unit(self, string, unit="Eh"):
        # if there are multiple numbers in the string, get the one
        # labeled with unit otherwise get the last number
        stringList = string.split()
        if unit in string:
            index = stringList.index(unit)
            return float(stringList[index - 1])
        else:
            return float(stringList[-1])

    def _get_energy_type(self, search="FINAL SINGLE POINT ENERGY", index=-1, unit="a.u.", filterDotDotDot=True):
        # hacky way of getting rid of some entries: first get all, remove
        # those that are '... done'
        s = self.grep_output(search)
        if filterDotDotDot:
            s = [item for item in s if "..." not in item]
        s = s[index]
        if not isinstance(index, slice):
            return Units.convert(self._get_energy_correct_unit(s), "a.u.", unit)
        else:
            return [Units.convert(self._get_energy_correct_unit(x), "a.u.", unit) for x in s]

    def get_scf_iterations(self, index=-1):
        """Returns Number of SCF Iterations from the Output File.

        Set ``index`` to choose the n-th occurence, *e.g.* to choose
        an certain step. Also supports slices.
        Defaults to the last occurence.
        """
        s = self.grep_output("SCF CONVERGED AFTER")
        n = [int(x.split()[-3]) for x in s]
        return n[index]

    def get_energy(self, index=-1, unit="a.u."):
        """Returns 'FINAL SINGLE POINT ENERGY:' from the output file.

        Set ``index`` to choose the n-th occurence of the total energy
        in the output, *e.g.* to choose an certain step.
        Also supports slices.
        Defaults to the last occurence.
        """
        return self._get_energy_type("FINAL SINGLE POINT ENERGY", index=index, unit=unit)

    def get_dispersion(self, index=-1, unit="a.u."):
        """Returns 'Dispersion correction' from the output file.

        Set ``index`` to choose the n-th occurence of the dispersion energy
        in the output, *e.g.* to choose a certain step.
        Also supports slices.
        Defaults to the last occurence.
        """
        return self._get_energy_type("Dispersion correction", index=index, unit=unit)

    def get_zpe(self, index=-1, unit="a.u."):
        """Returns 'Non-thermal (ZPE) correction' from the output file.

        Set ``index`` to choose the n-th occurence of the ZPE in the output,
        *e.g.* to choose a certain step.
        Also supports slices.
        Defaults to the last occurence.
        """
        return self._get_energy_type("Non-thermal (ZPE) correction", index=index, unit=unit)

    def get_inner_energy(self, index=-1, unit="a.u."):
        """Returns 'Total thermal energy' from the output file.

        Set ``index`` to choose the n-th occurence of the inner energy in
        the output, *e.g.* to choose a certain step.
        Also supports slices.
        Defaults to the last occurence.
        """
        return self._get_energy_type("Total thermal energy", index=index, unit=unit)

    def get_enthalpy(self, index=-1, unit="a.u."):
        """Returns 'Total Enthalpy' from the output file.

        Set ``index`` to choose the n-th occurence of the enthalpy in the
        output, *e.g.* to choose a certain step.
        Also supports slices.
        Defaults to the last occurence.
        """
        return self._get_energy_type("Total Enthalpy", index=index, unit=unit, filterDotDotDot=False)

    def get_entropy(self, index=-1, unit="a.u."):
        """Returns 'Final entropy term' from the output file.

        Set ``index`` to choose the n-th occurence of the entropy in the
        output, *e.g.* to choose a certain step.
        Also supports slices.
        Defaults to the last occurence.
        """
        return self._get_energy_type("Final entropy term", index=index, unit=unit, filterDotDotDot=False)

    def get_gibbs_free_energy(self, index=-1, unit="a.u."):
        """Returns 'Final Gibbs free energy' from the output file.

        Set ``index`` to choose the n-th occurence of the Gibbs free energy in
        the output, *e.g.* to choose a certain step.
        Also supports slices.
        Defaults to the last occurence.
        """
        return self._get_energy_type("Final Gibbs free energy", index=index, unit=unit, filterDotDotDot=False)

    def get_electrons(self, index=-1, spin_resolved=False):
        """Get Electron count from Output.

        Set ``spins`` to ``True`` to get a tuple of alpha and beta electrons
        instead of totals.  Set ``index`` to choose the n-th occurcence in the
        output, *e.g.* to choose a certain step.
        Also supports slices.
        Defaults to the last occurence.
        """
        if spin_resolved:
            alpha = [float(s.split()[-2]) for s in self.grep_output("N(Alpha)")][index]
            beta = [float(s.split()[-2]) for s in self.grep_output("N(Beta)")][index]
            if isinstance(alpha, float):
                alpha = [alpha]
                beta = [beta]
            ret = [(a, b) for a, b in zip(alpha, beta)]
        else:
            ret = [float(s.split()[-2]) for s in self.grep_output("N(Total)")][index]
        return ret

    def get_gradients(self, match=0, energy_unit="a.u.", dist_unit="bohr"):
        """Returns list of ndarrays with forces from the output
        (there the unit is a.u./bohr).

        ``match`` is passed to :meth:`~Results.get_output_chunk`,
        defaults to 0.
        """
        conv = Units.conversion_ratio("a.u.", energy_unit) / Units.conversion_ratio("bohr", dist_unit)
        searchBegin = "CARTESIAN GRADIENT"
        searchEnd = "Difference to translation invariance:"
        block = self.get_output_chunk(begin=searchBegin, end=searchEnd, match=match)
        ret = []
        for line in block:
            line = line.strip().split()
            if len(line) == 1:  # ---- lines means new set of forces
                ret.append([])
                continue
            if not line:  # ignore empty line
                continue
            ret[-1].append(line[-3:])
        ret = [np.array(item, dtype=float) * conv for item in ret]
        if len(ret) == 1:
            return ret[0]
        else:
            return ret

    def get_dipole_vector(self, index=-1, unit="a.u."):
        """Get the Dipole Vector
        Returns the dipole vector, expressed in *unit*.
        """
        lines = self.grep_output("Total Dipole Moment")
        conv = Units.conversion_ratio("a.u.", unit)
        vectors = [np.array(line.split()[-3:], dtype=float) * conv for line in lines]
        return vectors[index]

    def get_dipole(self, **kwargs):
        """Get Hirshfeld Analysis from Output

        Uses :meth:`~ORCAResults.get_dipole_vector` to calculate the
        total dipole.
        All options are passed on.
        """
        vec = self.get_dipole_vector(**kwargs)
        if isinstance(vec, np.ndarray):
            vec = [vec]
        vec = np.linalg.norm(vec, axis=1)
        if len(vec) == 1:
            return vec[0]
        else:
            return vec

    def get_atomic_charges(self, method="mulliken", match=0):
        """Get Atomic Charges from Output

        :parameter method: Can be any one that is available in the output, e.g. mulliken or loewdin.
        :parameter match: Select occurence in the output to use. E.g. when running multiple structures at once. Is passed to :meth:`~Results.get_output_chunk`, defaults to 0.
        """
        block = self.get_output_chunk(begin="{} ATOMIC CHARGES".format(method.upper()), end="-" * 25, match=match)
        ret = []
        for line in block:
            line = line.strip().split()
            if len(line) == 1:  # new set of charges
                ret.append([])
                continue
            if ("Sum" in line) or (not line):  # empty and sum lines
                continue
            ret[-1].append(float(line[-1]))
        if len(ret) == 1:
            return ret[0]
        else:
            return ret

    def get_hirshfeld(self, return_spin=False, match=0, skip=5):
        """Get Hirshfeld Analysis from Output

        :parameter return_spin: Return a tuple of (charge, spin) instead of just charge.
        :parameter match: Select occurence in the output to use. E.g. when running multiple structures at once. Is passed to :meth:`~Results.get_output_chunk`, defaults to 0.
        :parameter skip: Number of lines after the keyword in the outputfile to be skipped. Don't touch if you don't have trouble with your ORCA versions output.
        """
        block = self.get_output_chunk(begin="HIRSHFELD ANALYSIS", end="TOTAL", match=match)
        ret = []
        j = 0
        for i in range(len(block)):
            line = block[j].strip().split()
            if len(line) == 1:  # new set of charges
                ret.append([])
                j += skip
            elif line:
                if return_spin:
                    ret[-1].append(tuple([float(x) for x in line[-2:]]))
                else:
                    ret[-1].append(float(line[-2]))
            j += 1
            if j >= len(block):
                break
        if len(ret) == 1:
            return ret[0]
        else:
            return ret

    def get_orbital_energies(self, unit="a.u.", return_occupancy=False, match=0):
        """Returns Orbital Energies.

        :parameter return_occupancy: Set to *True* to recieve a tuple (Energy, Occupation) for each MO.
        :parameter match: Select occurence in the output to use. E.g. when running multiple structures at once. Is passed to :meth:`~Results.get_output_chunk`, defaults to 0.
        """
        conv = Units.conversion_ratio("a.u.", unit)
        block = self.get_output_chunk(begin="  NO   OCC          E(Eh)            E(eV) ", end="--", match=match)
        ret = []
        fastFWD = False
        for line in block:
            line = line.strip().split()
            if (len(line) == 4) and (line[0] == "0"):  # new set of energies
                ret.append([])
                fastFWD = False
            elif not len(line) == 4:
                fastFWD = True
                continue
            elif fastFWD:
                continue

            if return_occupancy:
                ret[-1].append((float(line[-2]) * conv, float(line[-3])))
            else:
                ret[-1].append(float(line[-2]) * conv)
        if len(ret) == 1:
            return ret[0]
        else:
            return ret


class ORCAJob(SingleJob):
    """
    A class representing a single computational job with
    `ORCA <https://orcaforum.cec.mpg.de>`_.

    In addition to the arguments of |SingleJob|, |ORCAJob| takes a
    ``copy_files`` argument. ``copy_files`` can be a list or string,
    containing paths to files to be copied to the jobs directory.
    This might e.g. be a molecule, restart files etc. By setting
    ``copy_symlink``, the files are not copied, but symlinked with
    relative links. The same things can be passed using the ``settings``
    instance of the job, i.e. ``self.settings.copy_files`` and
    ``self.settings.copy_symlink``. The former overwrites the latter.
    """

    _result_type = ORCAResults

    def __init__(self, copy_files=None, copy_symlink=False, **kwargs):
        SingleJob.__init__(self, **kwargs)
        if copy_files:
            self.settings.copy_files = copy_files
        if copy_symlink:
            self.settings.copy_symlink = copy_symlink

    def _get_ready(self):
        """Copy files to execution dir if self.copy_files is set."""
        SingleJob._get_ready(self)
        if "copy_files" in self.settings:
            if not isinstance(self.settings.copy_files, list):
                copy_files = [self.settings.copy_files]
            else:
                copy_files = self.settings.copy_files
            for f in copy_files:
                if ("copy_symlink" in self.settings) and (self.settings.copy_symlink):
                    symlink(relpath(f, self.path), opj(self.path, basename(f)))
                else:
                    shutil.copy(f, self.path)
        return

    def get_input(self):
        """
        Transform all contents of ``input`` branch of  ``settings`` into
        string with blocks, subblocks, keys and values.
        """

        def get_end(s):
            if (not isinstance(s, Settings)) or ("_end" not in s):
                return s
            else:
                return "{} end".format(s["_end"])

        def pretty_print_inner(s, indent):
            inp = ""
            for i, (key, value) in enumerate(s.items()):
                end = get_end(value)
                if i == 0:
                    inp += " {} {}\n".format(key, end)
                else:
                    inp += "{}{} {}\n".format(indent, key, end)
            return inp

        def pretty_print_orca(s, indent="", print_main=False):
            """Set print_main to true for initial call to have main section
            at the top
            """
            inp = ""
            if print_main:
                inp += "! {}\n\n".format(pretty_print_orca(s.main))
                pretty_print_orca(s, indent)
            if isinstance(s, Settings):
                for k, v in s.items():
                    if k in ("main", "molecule"):
                        # skip the molecule and main section
                        continue
                    else:
                        indent2 = (len(k) + 2) * " "
                        if not isinstance(v, Settings):
                            inp += "%{} {}\n\n".format(k, pretty_print_orca(v))
                        else:
                            block = pretty_print_inner(v, indent2)
                            inp += "%{}{}{}end\n\n".format(k, block, indent)
            elif isinstance(s, list):
                inp += "{}{}".format(indent, " ".join(s))
            else:
                inp += "{}{}".format(indent, s)
            return inp

        inp = pretty_print_orca(self.settings.input, print_main=True)
        if "molecule" in self.settings.input:
            inp += "* {}\n".format(self.settings.input.molecule)
        elif isinstance(self.molecule, Molecule):
            inp += self.print_molecule()
        else:
            raise JobError("ORCA Job needs a molecule to run.")

        return inp

    def print_molecule(self):
        """Print a molecule in the ORCA format using the xyz notation."""
        mol = self.molecule
        if "charge" in mol.properties and isinstance(mol.properties.charge, int):
            charge = mol.properties.charge
        else:
            charge = 0
        if "multiplicity" in mol.properties and isinstance(mol.properties.multiplicity, int):
            multi = mol.properties.multiplicity
        else:
            multi = 1
        # very accurate but this is equal to the accuracy of the ORCA output
        xyz = "\n".join(at.str(symbol=True, space=21, decimal=14) for at in mol.atoms)
        return "* xyz {} {}\n{}\n*\n\n".format(charge, multi, xyz)

    def get_runscript(self):
        """Returned runscript is just one line:
        ``orca myinput.inp``
        """
        return "orca {}".format(self._filename("inp"))

    def check(self):
        """Look for the normal termination signal in the output."""
        s = self.results.grep_output("ORCA TERMINATED NORMALLY")
        return len(s) > 0



#!/usr/bin/env python

from weakref import WeakValueDictionary

from scm.plams.core.errors import PlamsError
from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.amsanalysis import AMSAnalysisJob
from scm.plams.tools.kftools import KFFile
from scm.plams.trajectories.rkffile import RKFTrajectoryFile
from scm.plams.trajectories.rkfhistoryfile import RKFHistoryFile

__all__ = ["Trajectory"]


class Trajectory:
    """
    Class representing an AMS trajectory

    It creates molecule objects along the trajectory in a list-like fashion.
    It can also perform analysis tasks.

    Basic use of the |Trajectory| class works as follows:

        >>> from scm.plams import Trajectory

        >>> trajec = Trajectory('ams.rkf')
        >>> nsteps = len(trajec)

        >>> mol100 = trajec[100]
        >>> for i,mol in enumerate(trajec):
        ...     if i == 100:
        ...         print (mol is mol100)
        ...
        True

    As with a regular list, accessing the same element twice yields the same instance of the molecule.

    A trajectory may consist of multiple RKF files, for instance after a restarted MD run.

        >>> trajec = Trajectory(['md1.results/ams.rkf','md2.results/ams.rkf'])
        >>> nsteps = len(trajec)
        >>> print ('NSteps in each file: ',trajec.lengths)
        NSteps in each file:  [181, 19]

    Otherwise the object behaves the same, and iteration occurs over the concatenated files.

    .. note::

            It is possible to change the molecule objects once they are loaded, but this will not change the underlying files.
            When the altered molecule objects have been garbage collected, and the same frame is read anew, this frame
            will have the original molecule attributes.

                >>> import gc
                >>> from scm.plams import Trajectory

                >>> trajec = Trajectory('ams.rkf')
                >>> mol = trajec[100]
                >>> print (mol.properties.charge)
                0.0
                >>> mol.properties.charge = 10000.
                >>> print (trajec[100].properties.charge)
                1000.0

                >>> # remove the reference
                >>> mol = None
                >>> gc.collect()
                >>> print (trajec[100].properties.charge)
                0.

    The |Trajectory| object can also be used to perform analysis, using |AMSAnalysisJob| behind the scenes.

        >>> from scm.plams import Settings, Trajectory
        >>> from scm.plams import init, finish

        >>> trajec = Trajectory('ams.rkf')

        >>> # Define the type of analysis
        >>> settings = Settings()
        >>> settings.input.Task = 'RadialDistribution'
        >>> settings.input.RadialDistribution.AtomsFrom.Element = ['O']
        >>> settings.input.RadialDistribution.AtomsTo.Element = ['O']

        >>> # Run the analysis
        >>> init()
        >>> plots = trajec.run_analysis(settings)
        >>> finish()

        >>> # Store the plot in human readable format
        >>> for plot in plots :
        ...     plot.write('%s'%(plot.name+'.txt'))

    This results in a text file named RadialDistribution_1.txt, which contains data organized in columns.
    """

    def __init__(self, filenames):
        """
        Initiates instance of an AMS trajectory object

        * ``filenames`` -- List of filepaths of RKF trajectory files or single filepath for RKF file

        .. note::

                The corresponding file needs to remain on disc while the object is alive

        """
        if isinstance(filenames, str):
            filenames = [filenames]

        # Ceck is RKFHistoryFile needs to be used
        classes = []
        for fn in filenames:
            if not fn.split(".")[-1] == "rkf":
                raise PlamsError("Files need to be RKF")
            kf = KFFile(fn)
            if kf.reader is None:
                raise PlamsError("RKF file %s not found" % (fn))
            kf.reader._create_index()
            if not "History" in kf.reader._sections.keys():
                raise PlamsError("%s is not a trajectory file" % (fn))
            if "SystemVersionHistory" in kf.reader._sections.keys():
                classes.append(RKFHistoryFile)
            else:
                classes.append(RKFTrajectoryFile)

        # Store the file objects and associated data for each file
        self.files = [classes[i](fn) for i, fn in enumerate(filenames)]
        self.molecules = [rkf.get_plamsmol() for rkf in self.files]
        self.lengths = [len(rkf) for rkf in self.files]

        # Store the items that have been read as a weak reference
        self.weakrefs = WeakValueDictionary()

    def __len__(self):
        """
        Magic method that returns the length of the trajectory
        """
        return sum(self.lengths)

    def __iter__(self):
        """
        Iterates over molecule objects
        """
        for irkf, rkf in enumerate(self.files):
            mol = self.molecules[irkf]
            for istep in range(self.lengths[irkf]):
                i = self._get_index(irkf, istep)
                if i in self.weakrefs.keys():
                    yield self.weakrefs[i]
                    continue
                crd, cell = rkf.read_frame(istep, molecule=mol)
                m = mol.copy()
                self.weakrefs[i] = m
                yield m

    def __getitem__(self, s):
        """
        Returns Molecule object

        * ``s`` -- Python slice object
        """
        is_slice = False
        if isinstance(s, int):
            s = slice(s)
            start, stop, step = s.indices(len(self))
            indices = range(stop, stop + 1)
        elif isinstance(s, slice):
            is_slice = True
            start, stop, step = s.indices(len(self))
            indices = range(start, stop, step)
        mols = []
        for i in indices:
            if i in self.weakrefs.keys():
                mol = self.weakrefs[i]
            else:
                irkf, istep = self._get_filenum_and_stepnum(i)
                # This is always the first molecule
                mol = self.molecules[irkf].copy()
                crd, cell = self.files[irkf].read_frame(istep, molecule=mol)
                self.weakrefs[i] = mol
            mols.append(mol)
        if not is_slice:
            mols = mols[0]
        return mols

    def run_analysis(self, settings, steprange=None):
        """
        Calls the AMS analysis tool behind the scenes

        * ``settings``  -- PLAMS Settings object
          Example :

            >>> settings = Settings()
            >>> settings.input.Task = 'AutoCorrelation'
            >>> settings.input.AutoCorrelation.Property = 'Velocities'
            >>> settings.input.AutoCorrelation.MaxFrame = 2000

        * ``steprange`` -- Start frame, end frame, and stepsize. The default (None) corresponds to all frames in the object

        Returns a list of AMSAnalysisPlot objects.
        Main attributes of the AMSAnalysisPlot objects are :

        * ``name``    -- The name of the plot
        * ``x``       -- A list of lists containing the values of the coordinate system
          If the coordinate system is 1D, then it is a list containing a single list of values
          x = [[x1,x2,x3,x4,...,xn]]
        * ``y``       -- A list containing the function values
        * ``write()`` -- A method returning a string containing all plot info
        """
        # First get all the indices
        s = slice(steprange)
        start, stop, step = s.indices(len(self))
        indices = range(start, stop, step)
        if steprange is not None:
            if len(steprange) == 1:
                indices = range(stop, stop + 1)
        stepnums = {}
        for i in indices:
            irkf, istep = self._get_filenum_and_stepnum(i)
            if not irkf in stepnums:
                stepnums[irkf] = []
            stepnums[irkf].append(istep)
        ranges = []
        for irkf, values in stepnums.items():
            ranges.append((values[0] + 1, values[-1] + 1, step))
        print("ranges: ", ranges)

        # Now completel the settings object
        trajecsettings = []
        for ikf, rkf in enumerate(self.files):
            s = Settings()
            s.Trajectory.KFFilename = self.files[0].file_object.path
            s.Trajectory.Range = "%i %i %i" % (ranges[irkf][0], ranges[irkf][1], ranges[irkf][2])
            trajecsettings.append(s)
        settings.input.TrajectoryInfo = trajecsettings

        # Run the analysis job and extract the plots
        job = AMSAnalysisJob(settings=settings)
        result = job.run()
        plots = result.get_all_plots()
        return plots

    #################
    # Private methods
    #################

    def _get_filenum_and_stepnum(self, i):
        """
        Connects a filenumber and a stepnumber to index i
        """
        irkf, istep = (-1, -1)
        counter = 0
        for il, length in enumerate(self.lengths):
            if i < counter + length:
                irkf = il
                istep = i - counter
                break
            counter += length
        if irkf < 0:
            raise KeyError("Index out of range")
        return irkf, istep

    def _get_index(self, irkf, istep):
        """
        Connects the index i to a filenumber and a stepnumber
        """
        return sum([l for l in self.lengths[:irkf]]) + istep


def make_recursively_immutable(obj, verbose=False):
    """
    Return a new object that is immutable

    Note: Any recursive objects are NOT affected. For example, if obj is an instance of Molecule,
          then this function will return an instance of _Molecule, but new_obj.atoms[0].mol is still the original Molecule.
    """
    # Find all the branches of this object tree
    branchlist = [branch for branch in get_branches(obj)]
    if verbose:
        print("Tree: ")
        for branch in branchlist:
            print([n[0] for n in branch])

    # Find length of longest branch
    maxlength = max([len(branch) for branch in branchlist])
    for i in range(maxlength - 1, 0, -1):
        if verbose:
            print("Level: ", i)
        # Find all objects at this level, and theit parent
        for branch in branchlist:
            if len(branch) > i:
                current_obj = branch[i][1]
                current_name = branch[i][0]
                # Now make immutable. Take care to do this only once for an object
                if current_obj.__class__.__name__[0] == "_" or isinstance(current_obj, tuple):
                    continue
                current_obj = make_immutable(current_obj)
                parent_obj = branch[i - 1][1]
                if isinstance(parent_obj, list):
                    parent_obj[current_name] = current_obj
                elif isinstance(parent_obj, dict):
                    parent_obj[current_name] = current_obj
                else:
                    parent_obj.__dict__[current_name] = current_obj

    # Now do the top layer
    current_obj = make_immutable(obj, add_self=True)
    return current_obj


def make_immutable(obj, add_self=False):
    """
    Make this object immutable

    Note: This is specifically written for a Molecule object
    """

    def trytosetattr(self, name, value):
        """
        A setattr function that does not write attributes
        """
        # raise AttributeError("Cannot change object")
        if hasattr(self, name):
            raise AttributeError("Cannot reassign members")
        self.__dict__[name] = value

    def trytosetitem(self, name, value):
        """
        A setitem function for a dictionary that returns an exception
        """
        raise AttributeError("Cannot change object")

    def mutable_copy(obj, atoms=None):
        """
        Returns a copy of self, but one that is mutable
        """
        if hasattr(obj, "__dict__"):
            if "_orig" in obj.__dict__:
                return obj._orig
            else:
                return None
        elif "_orig" in obj:
            return obj["_orig"]
        else:
            return None

    # If the object is a built-in object, all the below does not work
    if isinstance(obj, list):
        return tuple(obj)
    # There will be dictionaries (Settings objects). How to make those immutable?

    # Lets get the name of the class
    classname = obj.__class__.__name__

    # Lets get all the instance variables of the input object
    attribs = []
    values = []
    if hasattr(obj, "__dict__"):
        attribs = [key for key in obj.__dict__.keys()]
        values = [obj.__dict__[key] for key in attribs]

    # Create a new class which is derived from the objects class
    ImmutableClass = type("_" + classname, (obj.__class__,), {})

    # Now create an instance of the new class with the same instance variables
    im_obj = ImmutableClass.__new__(ImmutableClass)
    if isinstance(obj, dict):
        # This is for a dictionary or PLAMS Settings object
        for key, value in obj.items():
            im_obj[key] = value
    else:
        for key, value in zip(attribs, values):
            setattr(im_obj, key, value)

    # Add its original self, for copying purposes
    if add_self:
        im_obj._orig = obj
        setattr(ImmutableClass, "copy", mutable_copy)

    # Now set the method that makes overwriting attributes impossible
    if isinstance(obj, dict):
        setattr(ImmutableClass, "__setitem__", trytosetitem)
    else:
        setattr(ImmutableClass, "__setattr__", trytosetattr)

    return im_obj


def get_branches(obj, subpath=None):
    """
    Search recursively through instance variables of object until an immutable one is met
    """
    tree = dictionary_from_object(obj)
    if subpath is None:
        subpath = (("top", obj),)
    # Identify immutable objects and circular objects (they are leafs)
    prev_objects = [node[1] for node in subpath]
    im_objs = []
    for n, o in tree.items():
        if is_immutable(o):
            im_objs.append(n)
        # Avoid circular crap
        elif True in [o is po for po in prev_objects]:
            im_objs.append(n)
    # The branch ends if the tree contains only immutable objects
    if len(im_objs) == len(tree):
        yield subpath
    else:
        for n, o in tree.items():
            if n in im_objs:
                continue
            for path in get_branches(o, subpath + ((n, o),)):
                yield path


def is_immutable(obj):
    """
    Check if this object is immutable

    Note: This is a first incomplete implementation
    """
    typelist = [int, float, tuple, str, bool]  # This does not take into account numpy integers etc
    immutable = False
    if obj is None:
        immutable = True
    for t in typelist:
        if isinstance(obj, t):
            immutable = True
            break
    return immutable


def dictionary_from_object(obj):
    """
    Get the dicrtionary of object attributes
    """
    if hasattr(obj, "__dict__"):
        tree = obj.__dict__
    elif isinstance(obj, dict):
        tree = obj
    elif hasattr(obj, "__iter__"):
        tree = {}
        for i, v in enumerate(obj):
            tree[i] = v
    else:
        raise Exception("What is this object?", type(obj))
    return tree



import pytest

from scm.plams import Molecule


@pytest.fixture
def m1(xyz_folder):
    return Molecule(xyz_folder / "CO_flat4_1.xyz")


@pytest.fixture
def m2(xyz_folder):
    return Molecule(xyz_folder / "CO_flat4_2.xyz")


@pytest.fixture
def m3(xyz_folder):
    return Molecule(xyz_folder / "CO_flat4_3.xyz")


@pytest.fixture
def m4(xyz_folder):
    return Molecule(xyz_folder / "CO_flat4_4.xyz")


def testYES(m1, m2, m3, m4):
    for i in range(4):
        assert m1.label(i) == m2.label(i)
    for i in range(4):
        assert m3.label(i) == m4.label(i)


def testNO(m1, m2, m3, m4):
    assert m1.label(4) != m2.label(4)
    assert m3.label(4) != m4.label(4)



import numpy as np
from typing import Union, Literal, Tuple


def _gaussian(x, A, x0, sigma):
    return A * np.exp(-((x - x0) ** 2) / (2 * sigma**2))


def _lorentzian(x, A, x0, sigma):
    return (A / np.pi) * (0.5 * sigma) / ((x - x0) ** 2 + (0.5 * sigma) ** 2)


def _generate_broadening_widths(x_data, broadening_width: Union[float, np.ndarray], centers):
    if not isinstance(broadening_width, np.ndarray):
        sigmas = x_data * 0 + broadening_width
    else:
        if len(broadening_width) == len(centers):
            sigmas = broadening_width
        else:
            raise ValueError(f"{len(broadening_width)=} != {len(centers)=} they should be equal")
    return sigmas


def broaden_results(
    centers: np.ndarray,
    areas: np.ndarray,
    broadening_width: Union[float, np.ndarray] = 40,
    broadening_type: Literal["gaussian", "lorentzian"] = "gaussian",
    x_data: Union[np.ndarray, Tuple[float, float, float]] = (0, 4000, 0.5),
    post_process: Literal["max_to_1"] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """convenient function for create xy curve with gaussian or lorentzian peaks. Used to obtain IR or Raman spectra for example.

    :param centers: x positions of the centers of the peaks
    :type centers: np.ndarray
    :param areas: areas of the peaks
    :type areas: np.ndarray
    :param broadening_width: if is np.ndarray each peak broadening is assigned otherwise apply the same value for all the peaks, defaults to 40
    :type broadening_width: Union[float, np.ndarray], optional
    :param broadening_type: the line shape of the peak, defaults to "gaussian"
    :type broadening_type: Literal['gaussian', 'lorentzian'], optional
    :param x_data: it can be a np.ndarray or you can set a tuple with (min,max,step_size), defaults to (0, 4000, 0.5)
    :type x_data: Union[np.ndarray, Tuple[float, float, float]], optional
    :param post_process: if max_to_1 the resulted spectrum has the max peak hight =1, defaults to None
    :type post_process: Literal['max_to_1'], optional
    :return: the x and y arrays
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    Function_Broaden = {
        "gaussian": _gaussian,
        "lorentzian": _lorentzian,
    }
    if isinstance(x_data, tuple) and len(x_data) == 3:
        x_data = np.arange(*x_data)

    y_data: np.ndarray = x_data * 0
    sigmas = _generate_broadening_widths(x_data, broadening_width, centers)

    for freq_i, Abs_i, sigma_i in zip(centers, areas, sigmas):
        y_data += Function_Broaden[broadening_type](x_data, A=Abs_i, x0=freq_i, sigma=sigma_i)

    if post_process == "max_to_1":
        y_data /= np.max(y_data)
    return x_data, y_data



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from scm.plams.mol.molecule import Molecule


class AsArrayContext:
    """A context manager for temporary inter-converting between PLAMS molecules and numpy arrays."""

    def __init__(self, mol: Molecule):
        self._atoms = mol.atoms
        self._from_array = mol.from_array

    def __call__(self, atom_subset=None):
        """Return cartesian coordinates of this molecule's atoms as a numpy array.

        *atom_subset* argument can be used to specify only a subset of atoms, it should be an iterable container with atoms belonging to this molecule.

        Returned value is a n*3 numpy array where n is the number of atoms in the whole molecule, or in *atom_subset*, if used.

        Alternatively, this function can be used in conjunction with the ``with`` statement,
        which automatically calls :meth:`Molecule.from_array` upon exiting the context manager.
        Note that the molecules' coordinates will be updated based on the array that was originally returned,
        so creating and operating on a copy thereof will not affect the original molecule.

        .. code-block:: python

            >>> from scm.plams import Molecule
            >>> mol = Molecule(...)
            >>> with mol.as_array as xyz_array:
            >>>     xyz_array += 5.0
            >>>     xyz_array[0] = [0, 0, 0]
            # Or equivalently
            >>> xyz_array = mol.as_array()
            >>> xyz_array += 5.0
            >>> xyz_array[0] = [0, 0, 0]
            >>> mol.from_array(xyz_array)
        """
        atom_subset = atom_subset or self._atoms

        try:
            at_len = len(atom_subset)
        except TypeError:  # atom_subset is an iterator
            count = -1
            shape = -1, 3
        else:
            count = at_len * 3
            shape = at_len, 3

        atom_iterator = itertools.chain.from_iterable(at.coords for at in atom_subset)
        xyz_array = np.fromiter(atom_iterator, count=count, dtype=float)
        xyz_array.shape = shape
        return xyz_array

    def __enter__(self, atom_subset=None):
        """Enter the context manager; return the Cartesian coordinate array."""
        self._atom_subset = atom_subset
        self._xyz_array = self.__call__(atom_subset=atom_subset)
        return self._xyz_array

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager; update the ``mol`` coordinates with those from the Cartesian coordinate array."""
        self._from_array(self._xyz_array, self._atom_subset)



import builtins
import pytest
import os

real_import = builtins.__import__


def get_mock_import_function(import_error_name):
    """
    Gets a mock version of the import function, which will raise an ImportError for the specified module,
    but import all other modules as usual.
    """

    def import_with_error(name, globals=None, locals=None, fromlist=(), level=0):
        if name == import_error_name:
            raise ImportError(f"Mock import error for '{import_error_name}'")
        return real_import(name, globals=globals, locals=locals, fromlist=fromlist, level=level)

    return import_with_error


def skip_if_no_ams_installation():
    """
    Check whether the AMSBIN environment variable is set, and therefore if there is an AMS installation present.
    If there is no installation, skip the test with a warning.
    """
    if os.getenv("AMSBIN") is None:
        pytest.skip("Skipping test as cannot find AMS installation. '$AMSBIN' environment variable is not set.")



from scm.plams.core.errors import FileError, PlamsError

__all__ = ["PDBRecord", "PDBHandler"]


_multiline = set(
    [
        "AUTHOR",
        "CAVEAT",
        "COMPND",
        "EXPDTA",
        "MDLTYP",
        "KEYWDS",
        "SOURCE",
        "SPLIT ",
        "SPRSDE",
        "TITLE ",
        "FORMUL",
        "HETNAM",
        "HETSYN",
        "SEQRES",
        "SITE  ",
        "REMARK",
    ]
)

_sequence = [
    "HEADER",
    "OBSLTE",
    "TITLE ",
    "SPLIT ",
    "CAVEAT",
    "COMPND",
    "SOURCE",
    "KEYWDS",
    "EXPDTA",
    "NUMMDL",
    "MDLTYP",
    "AUTHOR",
    "REVDAT",
    "SPRSDE",
    "JRNL  ",
    "REMARK",
    "DBREF ",
    "DBREF1",
    "DBREF2",
    "SEQADV",
    "SEQRES",
    "MODRES",
    "HET   ",
    "HETNAM",
    "HETSYN",
    "FORMUL",
    "HELIX ",
    "SHEET ",
    "SSBOND",
    "LINK  ",
    "CISPEP",
    "SITE  ",
    "CRYST1",
    "ORIGX1",
    "ORIGX2",
    "ORIGX3",
    "SCALE1",
    "SCALE2",
    "SCALE3",
    "MTRIX1",
    "MTRIX2",
    "MTRIX3",
    "MODEL ",
    "CONECT",
    "MASTER",
    "END   ",
]

_coord = ["ATOM  ", "ANISOU", "HETATM", "TER   ", "ENDMDL"]


# ===========================================================================


class PDBRecord:
    __slots__ = ["name", "value", "model"]

    def __init__(self, s):
        s = s.rstrip("\n")
        if len(s) < 80:
            s = "%-80s" % s
        self.name = s[:6]
        self.value = [s[6:]]
        self.model = []

    def __str__(self):
        res = ""
        if self.name != "_model":
            for val in self.value:
                res += self.name + " " + val + "\n"
        for i in self.model:
            res += str(i)
        return res

    def is_multiline(self):
        return self.name in _multiline

    def extend(self, s):
        s = s.rstrip("\n")

        def _tonum(ss):
            if ss.isspace():
                return 1
            else:
                return int(ss)

        if self.is_multiline() and self.name == s[:6]:
            val = s[6:]
            if self.name == "REMARK":
                self.value.append(val)
                return True
            beg, end = 1, 4
            if self.name == "FORMUL":
                beg, end = 10, 12
            last = _tonum(self.value[-1][beg:end])
            new = _tonum(val[beg:end])
            if new == last + 1:
                self.value.append(val)
                return True
        return False


# ===========================================================================


class PDBHandler:
    def __init__(self, textfile=None):
        self.records = {}
        for key in _sequence + _coord:
            self.records[key] = []
        if textfile is not None:
            if isinstance(textfile, str):
                try:
                    f = open(textfile, "r")
                except:
                    raise FileError("PDBHandler: Error reading file %s" % textfile)
                self.read(f)
                f.close()
            else:  # textfile is an open file object
                self.read(textfile)

    def singlemodel(self):
        return "_model" in self.records

    def read(self, f):
        model = None
        line = f.readline()
        while line:
            newrecord = PDBRecord(line)
            line = f.readline()
            while newrecord.extend(line):
                line = f.readline()
            key = newrecord.name
            self.records[key].append(newrecord)

            if model is not None and key in _coord:
                model.append(newrecord)

            if key == "MODEL ":
                model = newrecord.model
            if model is None and key in _coord[:3]:
                tmp = PDBRecord("_model")
                model = tmp.model
                model.append(newrecord)
                self.records["_model"] = tmp
            elif key == "END   ":
                break

    def write(self, f):
        for key in _sequence:
            if key == "MODEL " and self.singlemodel():
                f.write(str(self.records["_model"]))
            else:
                for record in self.records[key]:
                    f.write(str(record))

    def calc_master(self):
        def total(key):
            if key in _multiline:
                return sum([len(x.value) for x in self.records[key]])
            else:
                return len(self.records[key])

        remark = total("REMARK")
        het = total("HET   ")
        helix = total("HELIX ")
        sheet = total("SHEET ")
        site = total("SITE  ")
        conect = total("CONECT")
        seqres = total("SEQRES")
        xform = sum(
            map(total, ["ORIGX1", "ORIGX2", "ORIGX3", "SCALE1", "SCALE2", "SCALE3", "MTRIX1", "MTRIX2", "MTRIX3"])
        )

        if self.singlemodel():
            ter = total("TER   ")
            coord = total("ATOM  ") + total("HETATM")
        else:
            lst = self.records["MODEL "][0].model
            ter, coord = 0, 0
            for i in lst:
                if i.name == "TER   ":
                    ter += 1
                elif i.name in ["ATOM  ", "HETATM"]:
                    coord += 1

        master = "MASTER    %5i%5i%5i%5i%5i%5i%5i%5i%5i%5i%5i%5i          \n" % (
            remark,
            0,
            het,
            helix,
            sheet,
            0,
            site,
            xform,
            coord,
            ter,
            conect,
            seqres,
        )
        return PDBRecord(master)

    def check_master(self):
        if self.records["MASTER"]:
            old = self.records["MASTER"][0]
            new = self.calc_master()
            return old.value == new.value
        return False

    def get_models(self):
        if self.singlemodel():
            return [self.records["_model"].model]
        else:
            return [x.model for x in self.records["MODEL "]]

    def add_record(self, record):
        if record.name in self.records:
            self.records[record.name].append(record)
        elif record.name == "_model":
            self.records[record.name] = record
        else:
            raise PlamsError("PDBHandler.add_record: Invalid record passed")

    def add_model(self, model):
        # model: list of PDBRecords of type in _coord
        if "_model" in self.records:
            old = self.records["_model"]
            del self.records["_model"]

            newmodel = PDBRecord("MODEL     %4i" % 1)
            newmodel.model = old.model
            endmdl = PDBRecord("ENDMDL")
            newmodel.model.append(endmdl)
            self.add_record(newmodel)
            self.add_record(endmdl)

        if self.records["MODEL "]:  # there were 1+ models present before
            newmodel = PDBRecord("MODEL     %4i" % (1 + len(self.records["MODEL "])))
            newmodel.model = model
            if newmodel.model[-1].name != "ENDMDL":
                newmodel.model.append(PDBRecord("ENDMDL"))
            self.add_record(newmodel)
            self.records["NUMMDL"] = [PDBRecord("NUMMDL    %4i" % len(self.records["MODEL "]))]
        else:
            newmodel = PDBRecord("_model")
            newmodel.model = model
            self.add_record(newmodel)

        for rec in model:
            self.add_record(rec)



import sys
import os
import numpy
from scm.plams import AMSJob, Settings, Units, AMSAnalysisJob
from scm.plams import to_rdmol, init, finish


def main(filename):
    """
    Main functionality (charge assignment to ions)
    """
    # Get the molecular system
    job = AMSJob.load_external(filename)
    mol = job.molecule
    elements = [at.symbol for at in mol.atoms]

    # Assign atomic charges
    charges = mol.guess_atomic_charges(adjust_to_systemcharge=False)
    charges = add_metal_charges(mol, charges)

    # Find ions
    molindices = mol.get_molecule_indices()
    ions = {}
    ioncharges = {}
    nions = {}
    formulas = {}
    for imol, atoms in enumerate(molindices):
        submol = mol.get_fragment(atoms)
        label = submol.label()
        q = sum([charges[i] for i in atoms])
        if abs(q) > 1e-10:
            if not label in ions.keys():
                ions[label] = []
                ioncharges[label] = q
                nions[label] = 0
                formulas[label] = submol.get_formula()
            ions[label] += atoms
            nions[label] += 1
    for k, q in ioncharges.items():
        print(formulas[k], q, nions[k])

    return ions, ioncharges


def add_metal_charges(mol, charges):
    """
    Assign charges to the metal ions
    """
    system_charge = mol.properties.charge if not isinstance(mol.properties.charge, Settings) else 0.0
    if system_charge == sum(charges):
        return charges
    dq = system_charge - sum(charges)
    # print ('Charge difference: ',dq)

    elements = [at.symbol for at in mol.atoms]
    metals = [i for i, at in enumerate(mol.atoms) if at.is_metallic]
    # print ('Metals: ',metals)

    electron_affinities, ionization_energies = read_ea_ie()
    chi0 = {iat: 0.5 * (electron_affinities[elements[iat]] + ionization_energies[elements[iat]]) for iat in metals}
    J0 = {iat: ionization_energies[elements[iat]] - electron_affinities[elements[iat]] for iat in metals}
    Jc = 1.0

    # Set up the matrix (A) and vector b, to then solve Ax = b
    # The last element of the vector (and last row of A) hold the constraint sum(x_i) = Q_tot
    matrix = numpy.zeros((len(metals) + 1, len(metals) + 1))
    b = numpy.ones(len(metals) + 1)
    for i, iat in enumerate(metals):
        matrix[i, i] = J0[iat]
        b[i] = chi0[iat]
    matrix[-1, :-1] = 1.0
    matrix[:-1, -1] = 1.0
    b[-1] = dq
    # print ('Matrix:')
    # print (matrix)
    # print ('b: ',b)

    # Solve Ax+b
    mcharges = numpy.linalg.solve(matrix, b)
    mcharges_int = [round(q) for q in mcharges]
    if sum(mcharges_int) != dq:
        print([(elements[iat], mcharges[i]) for i, iat in enumerate(metals)])
        raise Exception("Predicted charges non-integer!")
    for i, iat in enumerate(metals):
        charges[iat] = mcharges_int[i]
    return charges


def read_ea_ie():
    """
    Read in the electron affinities and ionization energies from file
    """
    electron_affinities = {}
    ionization_energies = {}
    for el, l in data.items():
        electron_affinities[el] = data[el][0]
        ionization_energies[el] = data[el][1]
    return electron_affinities, ionization_energies


# Element:[     ElectronAffinity,     IonizationEnergy]
data = {
    "Li": [0.0227110811, 0.1981523452],
    "Na": [0.0201386286, 0.1888547667],
    "Al": [0.0162064511, 0.2199814425],
    "Si": [0.0508978112, 0.2995804744],
    "Sc": [0.0069088726, 0.2411123028],
    "Ti": [0.0029031965, 0.2509243718],
    "V": [0.0192933941, 0.2479109274],
    "Cr": [0.0244750486, 0.2486826632],
    "Fe": [0.0059901395, 0.2903931438],
    "Co": [0.0242913020, 0.2896214081],
    "Ni": [0.0424822164, 0.2807648214],
    "Cu": [0.0451281676, 0.2839252631],
    "Ga": [0.0110247967, 0.2204591837],
    "Ge": [0.0496115849, 0.2903196452],
    "Rb": [0.0171986828, 0.1535019187],
    "Y": [0.0112820419, 0.2284705360],
    "Zr": [0.0156552112, 0.2437950033],
    "Nb": [0.0328171447, 0.2483886686],
    "Mo": [0.0274149943, 0.2606261929],
    "Tc": [0.0202121272, 0.2675350654],
    "Ru": [0.0385867883, 0.2705117605],
    "Rh": [0.0417839793, 0.2741131941],
    "Pd": [0.0204693725, 0.3063790990],
    "Ag": [0.0478476175, 0.2784128648],
    "In": [0.0110247967, 0.2126315781],
    "Sn": [0.0440991866, 0.2698870221],
    "Sb": [0.0393217747, 0.3175141436],
    "Cs": [0.0173456801, 0.1431018606],
    "La": [0.0183746611, 0.2049509698],
    "Ce": [0.0183746611, 0.2035544955],
    "Ta": [0.0118332817, 0.2899521520],
    "W": [0.0299506976, 0.2932595910],
    "Re": [0.0055123983, 0.2895846587],
    "Os": [0.0404242544, 0.3197191029],
    "Ir": [0.0575126892, 0.3344188318],
    "Au": [0.0848541849, 0.3390492464],
    "Tl": [0.0073498644, 0.2244648598],
    "Pb": [0.0132297560, 0.2725697226],
    "Bi": [0.0347648588, 0.2678658093],
    "Po": [0.0698237121, 0.3093190448],
}

if __name__ == "__main__":
    if len(sys.argv) == 1:
        print("Usage: amspython get_chargee_ions.py path/to/ams.rkf")
        sys.exit(0)
    filename = sys.argv[1]

    ions, ioncharges = main(filename)

    outfile = open("charges.in", "w")
    for k, q in ioncharges.items():
        outfile.write("%8.1f: %s\n" % (q, " ".join(["%i" % (iat) for iat in ions[k]])))
    outfile.close()



from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.ams import AMSResults
import numpy as np
from scm.plams.recipes.md.amsmdjob import AMSNVTJob

__all__ = ["AMSMDScanDensityJob", "AMSMDScanDensityResults"]


class AMSMDScanDensityResults(AMSResults):
    """Results class for AMSMDScanDensityJob"""

    def get_lowest_energy_index(self, variable="Energy", history_section="History"):
        """
        Returns the 1-based index of the lowest energy molecule
        """
        energies = self.get_history_property(variable, history_section)
        minindex = np.argmin(energies) + 1
        return minindex

    def get_lowest_energy_molecule(self, variable="TotalEnergy"):
        return self.get_history_molecule(self.get_lowest_energy_index(variable, "MDHistory"))


class AMSMDScanDensityJob(AMSNVTJob):
    """A class for scanning the density using MD Deformations"""

    _result_type = AMSMDScanDensityResults

    def __init__(self, molecule, scan_density_upper=1.4, startstep=None, **kwargs):
        AMSNVTJob.__init__(self, molecule=molecule, **kwargs)

        from_density = self.molecule.get_density() * 1e-3
        orig_length = self.molecule.cell_lengths()
        density_ratio = from_density / scan_density_upper
        new_length = [x * density_ratio**0.333333 for x in orig_length]

        self.settings.input.ams.MolecularDynamics.NSteps

        self.scan_density_upper = scan_density_upper
        self.startstep = startstep or 1

        s = Settings()
        s.input.ams.MolecularDynamics.Deformation.TargetLength = " ".join([str(x) for x in new_length])
        s.input.ams.MolecularDynamics.Deformation.StartStep = self.startstep

        self.settings += s



import pytest

from scm.plams import Molecule


@pytest.fixture
def m1(xyz_folder):
    return Molecule(xyz_folder / "RS1.xyz")


@pytest.fixture
def m2(xyz_folder):
    return Molecule(xyz_folder / "RS2.xyz")


def testYES(m1, m2):
    for i in range(3):
        assert m1.label(i) == m2.label(i)


def testNO(m1, m2):
    for i in range(3, 5):
        assert m1.label(i) != m2.label(i)



from collections import OrderedDict

from scm.plams.core.basejob import MultiJob
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings
import numpy as np
from scm.plams.recipes.md.amsmdjob import AMSNVEJob

__all__ = ["AMSNVESpawnerJob", "AMSNVESpawnerResults"]


class AMSNVESpawnerResults(Results):
    """Results class for AMSNVESpawnerJob"""

    def get_all_dipole_derivatives_acf(
        self, start_fs=0, end_fs=None, every_fs=None, max_dt_fs=None, x=True, y=True, z=True, normalize=False
    ):
        ret_x, ret_y = [], []
        for job in self.job.children.values():
            xv, yv = job.results.get_dipole_derivatives_acf(
                start_fs=start_fs,
                end_fs=end_fs,
                every_fs=every_fs,
                max_dt_fs=max_dt_fs,
                x=x,
                y=y,
                z=z,
                normalize=normalize,
            )
            ret_x.append(xv)
            ret_y.append(yv)

        return ret_x, ret_y

    def get_dipole_derivatives_acf(
        self, start_fs=0, end_fs=None, every_fs=None, max_dt_fs=None, x=True, y=True, z=True, normalize=False
    ):
        all_x, all_y = self.get_all_dipole_derivatives_acf(
            start_fs=start_fs, end_fs=end_fs, every_fs=every_fs, max_dt_fs=max_dt_fs, x=x, y=y, z=z, normalize=normalize
        )
        return np.mean(all_x, axis=0), np.mean(all_y, axis=0)

    def get_mean_temperature(self) -> float:
        avg_T = np.mean([job.results.readrkf("MDResults", "MeanTemperature") for job in self.job.children.values()])

        return avg_T


class AMSNVESpawnerJob(MultiJob):
    """A class for running multiple NVE simulations with initial structures/velocities taken from an NVT trajectory. The NVT trajectory must contain the velocities!"""

    _result_type = AMSNVESpawnerResults

    def _default_settings(self):
        s = Settings()
        s.input.ForceField.Type = "GAFF"
        s.input.ForceField.AnteChamberIntegration = "Yes"
        return s

    def __init__(self, previous_job, n_nve=1, name="nvespawnerjob", **kwargs):  # needs to be finished
        """
        previous_job: AMSJob
            An AMSJob with an MD trajectory. Must contain velocities (WriteVelocities Yes). Note that the trajectory should have been equilibrated before it starts.

        n_nve : int
            The number of NVE simulations to spawn

        All other settings can be set as for an AMSNVEJob (e.g. ``nsteps``).

        """
        MultiJob.__init__(self, children=OrderedDict(), name=name)

        self.previous_job = previous_job
        self.n_nve = n_nve
        self.nve_constructor_settings = kwargs
        self.nve_jobs = []

    def prerun(self):  # noqa F811
        """
        Constructs the children jobs
        """

        # use previously run previous_job
        assert self.previous_job.status != "created", "You can only pass in a finished AMSJob"
        try:
            self.previous_job.results.readrkf("MDHistory", "Velocities(1)")
        except KeyError:
            raise KeyError("Couldn't read velocities from {}".format(self.previous_job.results.rkfpath()))

        nframes_in_history = self.previous_job.results.readrkf("History", "nEntries")

        if self.n_nve > 0:
            interval = nframes_in_history // self.n_nve
            frames = np.linspace(interval, nframes_in_history, self.n_nve, dtype=int)
            for i, frame in enumerate(frames):
                # the nve jobs only get different inside prerun, so need to do something to make them different to prevent the PLAMS rerun prevention
                name = f"nve{i+1}"
                self.settings.input.ams["#"] = name  # add a comment line

                self.children[name] = AMSNVEJob.restart_from(
                    self.previous_job, frame=frame, name=name, use_prerun=True, **self.nve_constructor_settings
                )
                self.children[name].from_frame = frame

                self.nve_jobs.append(self.children[name])



__all__ = [
    "PlamsError",
    "FileError",
    "ResultsError",
    "JobError",
    "PTError",
    "UnitsError",
    "MoleculeError",
    "TrajectoryError",
]


class PlamsError(Exception):
    """General PLAMS error."""


class FileError(PlamsError):
    """File or filesystem related error."""


class ResultsError(PlamsError):
    """|Results| related error."""


class JobError(PlamsError):
    """|Job| related error."""


class PTError(PlamsError):
    """:class:`Periodic table<scm.plams.utils.PeriodicTable>` error."""


class UnitsError(PlamsError):
    """:class:`Units converter<scm.plams.utils.Units>` error."""


class MoleculeError(PlamsError):
    """|Molecule| related error."""


class TrajectoryError(PlamsError):
    """:class:`Trajectory<scm.plams.trajectories.TrajectoryFile>` error."""
